import pluggy

from bluetooth_mesh.apps import hookspecs

hookimpl = pluggy.HookimplMarker("bluetooth_mesh")


def get_plugin_manager():
    plugin_manager = pluggy.PluginManager("bluetooth_mesh")
    plugin_manager.add_hookspecs(hookspecs)
    plugin_manager.load_setuptools_entrypoints("bluetooth_mesh")
    return plugin_manager



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
"""
This module implements D-Bus interfaces that need to be exposed by an
application using BlueZ mesh stack, described in detail in mesh-api.txt_.

They are not meant to be used directly. See :py:mod:`bluetooth_mesh.application` instead.

.. _mesh-api.txt: https://git.kernel.org/pub/scm/bluetooth/bluez.git/tree/doc/mesh-api.txt
"""
# pylint: disable=R0201,W0613,W0622

import asyncio
import logging
from collections import defaultdict
from datetime import timedelta
from typing import Any, Dict, Mapping, Optional, Tuple
from uuid import UUID

from dbus_next import DBusError, Variant
from dbus_next.service import PropertyAccess, ServiceInterface, dbus_property, method

from bluetooth_mesh.crypto import ApplicationKey, DeviceKey, NetworkKey


class DBusService:
    NAME = "org.freedesktop.DBus"
    PATH = "/org/freedesktop/DBus"


class DBusInterface:
    def __init__(self, dbus_service):
        self._interface = dbus_service.get_interface("org.freedesktop.DBus")

    async def get_name_owner(self, name):
        return await self._interface.call_get_name_owner(name)

    def on_name_owner_changed(self, callback):
        self._interface.on_name_owner_changed(callback)

    def off_name_owner_changed(self, callback):
        self._interface.off_name_owner_changed(callback)

    def on_name_lost(self, callback):
        self._interface.on_name_lost(callback)

    def off_name_lost(self, callback):
        self._interface.off_name_lost(callback)

    async def get_service(self, service):
        owner = asyncio.Future()

        def name_owner_changed(name, old_owner, new_owner):
            if name != service.NAME:
                return

            if not new_owner:
                return

            owner.set_result(new_owner)

        try:
            self.on_name_owner_changed(name_owner_changed)
            return await self.get_name_owner(service.NAME)
        except DBusError as ex:
            return await owner
        finally:
            self.off_name_owner_changed(name_owner_changed)


class MeshService:
    NAME = "org.bluez.mesh"
    PATH = "/org/bluez/mesh"


def extract_model_config(configuration):
    vendor_id = None
    model_conf_dict = {}

    for key, variant in configuration.items():
        if key == "Vendor":
            vendor_id = variant.value
            continue

        if key == "Bindings":
            model_conf_dict["bindings"] = variant.value
            continue

        if key == "PublicationPeriod":
            model_conf_dict["publication_period"] = timedelta(
                milliseconds=variant.value
            )
            continue

        if key == "Subscriptions":
            subs = []

            for addr in variant.value:
                if addr.signature == "q":
                    subs.append(addr.value)

                elif addr.signature == "ay":
                    subs.append(UUID(bytes=addr.value))

            model_conf_dict["subscriptions"] = set(subs)

    return vendor_id, model_conf_dict


class ElementInterface(ServiceInterface):
    def __init__(self, element):
        self.element = element
        super().__init__(name="org.bluez.mesh.Element1")

    @dbus_property(name="Models", access=PropertyAccess.READ)
    def get_models(self) -> "aq":
        return self.element.models

    @dbus_property(name="VendorModels", access=PropertyAccess.READ)
    def get_vendor_models(self) -> "a(qq)":
        return [list(model_id) for model_id in self.element.vendor_models]

    @method(name="MessageReceived")
    def message_received(
        self, source: "q", key_index: "q", destination: "v", data: "ay"
    ):
        if destination.signature == "q":
            destination = destination.value
        elif destination.signature == "ay":
            destination = UUID(destination.value)

        self.element.message_received(source, key_index, destination, data)

    @method(name="DevKeyMessageReceived")
    def dev_key_message_received(
        self, source: "q", remote: "b", net_index: "q", data: "ay"
    ):
        self.element.dev_key_message_received(source, remote, net_index, data)

    @method(name="UpdateModelConfiguration")
    def update_model_configuration(self, model_id: "q", configuration: "a{sv}"):
        vendor_id, model_config = extract_model_config(configuration)
        self.element.update_model_configuration((vendor_id, model_id), model_config)

    @dbus_property(name="Index", access=PropertyAccess.READ)
    def get_index(self) -> "y":
        return self.element.index

    @dbus_property(name="Location", access=PropertyAccess.READ)
    def get_location(self) -> "q":
        return self.element.location


class ProvisionAgentInterface(ServiceInterface):
    def __init__(self, application):
        self.application = application
        self.logger = logging.getLogger("ProvisionAgentInterface")
        super().__init__(name="org.bluez.mesh.ProvisionAgent1")

    @method(name="PrivateKey")
    def private_key(self) -> "ay":
        return self.application.private_key()

    @method(name="PublicKey")
    def public_key(self) -> "ay":
        return self.application.public_key()

    @method(name="DisplayString")
    def display_string(self, value: "s"):
        self.application.display_string(value)

    @method(name="DisplayNumeric")
    def display_numeric(self, type: "s", number: "u"):
        self.application.display_numeric(type, number)

    @method(name="PromptNumeric")
    def prompt_numeric(self, type: "s") -> "u":
        return self.application.prompt_numeric(type)

    @method(name="PromptStatic")
    def prompt_static(self, type: "s") -> "ay":
        return self.application.prompt_static(type)

    @method(name="Cancel")
    def cancel(self):
        pass

    @dbus_property(name="Capabilities", access=PropertyAccess.READ)
    def get_capabilities(self) -> "as":
        return [cap.value for cap in self.application.capabilities]

    @dbus_property(name="OutOfBandInfo", access=PropertyAccess.READ)
    def out_of_band_info(self) -> "as":
        return [oob.value for oob in self.application.oob_info]

    @dbus_property(name="URI", access=PropertyAccess.READ)
    def uri(self) -> "s":
        return self.application.uri


class ProvisionerInterface(ServiceInterface):
    def __init__(self, application):
        self.application = application
        self.logger = logging.getLogger("ProvisionerInterface")
        super().__init__(name="org.bluez.mesh.Provisioner1")

    @method(name="ScanResult")
    def scan_result(self, rssi: "n", data: "ay", options: "a{sv}"):
        self.application.scan_result(rssi, data, options)

    @method(name="RequestProvData")
    def request_prov_data(self, count: "y") -> "qq":
        return self.application.request_prov_data(count)

    @method(name="AddNodeComplete")
    def add_node_complete(self, uuid: "ay", unicast: "q", count: "y"):
        self.application.add_node_complete(uuid=uuid, unicast=unicast, count=count)

    @method(name="AddNodeFailed")
    def add_node_failed(self, uuid: "ay", reason: "s"):
        self.application.add_node_failed(uuid=uuid, reason=reason)


class _ApplicationInterface(ServiceInterface):
    def __init__(self, application):
        self.application = application
        self.logger = logging.getLogger("ApplicationInterface")
        super().__init__(name="org.bluez.mesh.Application1")

    @method(name="JoinComplete")
    async def join_complete(self, token: "t"):
        self.application.join_complete(token)

    @method(name="JoinFailed")
    async def join_failed(self, reason: "s"):
        self.application.join_failed(reason)

    @dbus_property(name="CompanyID", access=PropertyAccess.READ)
    def get_company_id(self) -> "q":
        return self.application.company_id

    @dbus_property(name="ProductID", access=PropertyAccess.READ)
    def get_product_id(self) -> "q":
        return self.application.product_id

    @dbus_property(name="VersionID", access=PropertyAccess.READ)
    def get_version_id(self) -> "q":
        return self.application.version_id


class _ApplicationInterfaceCRPL(_ApplicationInterface):
    @dbus_property(name="CRPL", access=PropertyAccess.READ)
    def get_crpl(self) -> "q":
        return self.application.crpl


def ApplicationInterface(application):
    if application.crpl is not None:
        return _ApplicationInterfaceCRPL(application)
    else:
        return _ApplicationInterface(application)


class NetworkInterface:
    def __init__(self, mesh_service):
        self._interface = mesh_service.get_interface("org.bluez.mesh.Network1")

    async def join(self, app_defined_root: str, uuid: UUID) -> None:
        await self._interface.call_join(app_defined_root, uuid.bytes)

    async def cancel(self) -> None:
        await self._interface.call_cancel()

    async def attach(
        self, app_defined_root: str, token: int
    ) -> Tuple[str, Dict[int, Dict[Tuple[Optional[int], int], Mapping[str, Any]]]]:
        path, configuration = await self._interface.call_attach(app_defined_root, token)

        configuration_dict = defaultdict(dict)

        for element, element_conf in configuration:
            for model_id, model_conf in element_conf:
                vendor_id, model_conf_dict = extract_model_config(model_conf)

                configuration_dict[element][(vendor_id, model_id)] = model_conf_dict

        return path, configuration_dict

    async def leave(self, token: int) -> None:
        await self._interface.call_leave(token)

    async def create_network(self, app_root: str, uuid: UUID) -> None:
        return await self._interface.call_create_network(app_root, uuid.bytes)

    async def import_node(
        self,
        app_root: str,
        uuid: UUID,
        dev_key: DeviceKey,
        net_key: NetworkKey,
        net_index: int,
        flags: Mapping[str, Any],
        iv_index: int,
        unicast: int,
    ) -> None:

        await self._interface.call_import(
            app_root,
            uuid.bytes,
            dev_key.bytes,
            net_key.bytes,
            net_index,
            flags,
            iv_index,
            unicast,
        )


class AclInterface:
    def __init__(self, mesh_service):
        self._interface = mesh_service.get_interface(
            "org.bluez.mesh.AccessControlList1"
        )

    async def grant_access(
        self, uuid: UUID, dev_key: DeviceKey, net_key: NetworkKey
    ) -> int:
        return await self._interface.call_grant_access(uuid, dev_key, net_key)

    async def revoke_access(self, token: int):
        return await self._interface.call_revoke_access(token)


class NodeInterface:
    def __init__(self, node_service):
        self._interface = node_service.get_interface("org.bluez.mesh.Node1")

    async def send(
        self, element_path: str, destination: int, key_index: int, data: bytes
    ) -> None:
        await self._interface.call_send(element_path, destination, key_index, data)

    async def dev_key_send(
        self,
        element_path: str,
        destination: int,
        remote: bool,
        net_index: int,
        data: bytes,
    ) -> None:
        await self._interface.call_dev_key_send(
            element_path, destination, remote, net_index, data
        )

    async def add_net_key(
        self,
        element_path: str,
        destination: int,
        subnet_index: int,
        net_index: int,
        update: bool,
    ) -> None:
        await self._interface.call_add_net_key(
            element_path, destination, subnet_index, net_index, update
        )

    async def add_app_key(
        self,
        element_path: str,
        destination: int,
        app_index: int,
        net_index: int,
        update: bool,
    ) -> None:
        await self._interface.call_add_app_key(
            element_path, destination, app_index, net_index, update
        )

    async def publish(self, element_path: str, model: int, data: bytes) -> None:
        await self._interface.call_publish(element_path, model, data)

    async def vendor_publish(
        self, element_path: str, vendor: int, model_id: int, data: bytes
    ):
        await self._interface.call_vendor_publish(element_path, vendor, model_id, data)

    async def update_sequence_number(self, seq_nr: int) -> int:
        updated_seq_nr = await self._interface.call_update_sequence_number(seq_nr)
        if updated_seq_nr != seq_nr:
            raise ValueError("Cannot update sequence number")
        return updated_seq_nr

    async def features(self) -> Mapping[str, bool]:
        return await self._interface.get_features()

    async def beacon(self) -> bool:
        return await self._interface.get_beacon()

    async def iv_update(self) -> bool:
        return await self._interface.get_iv_update()

    async def iv_index(self) -> int:
        return await self._interface.get_iv_index()

    async def seconds_since_last_heard(self) -> int:
        return await self._interface.get_seconds_since_last_heard()

    async def address(self) -> int:
        return (await self._interface.get_addresses())[0]

    async def sequence_number(self) -> int:
        return await self._interface.get_sequence_number()


class ManagementInterface:
    def __init__(self, node_service):
        self._interface = node_service.get_interface("org.bluez.mesh.Management1")

    async def unprovisioned_scan(self, **kwargs) -> None:
        options = dict(Seconds=Variant("q", kwargs.get("seconds", 0)))
        await self._interface.call_unprovisioned_scan(options)

    async def unprovisioned_scan_cancel(self) -> None:
        await self._interface.call_unprovisioned_scan_cancel()

    async def add_node(self, uuid: UUID, **kwargs) -> None:
        options = dict(
            # TODO: populate with options when defined
        )
        await self._interface.call_add_node(uuid.bytes, options)

    async def create_subnet(self, net_index: int) -> None:
        await self._interface.call_create_subnet(net_index)

    async def import_subnet(self, net_index: int, net_key: NetworkKey) -> None:
        await self._interface.call_import_subnet(net_index, net_key.bytes)

    async def update_subnet(self, net_index: int) -> None:
        await self._interface.call_update_subnet(net_index)

    async def delete_subnet(self, net_index: int) -> None:
        await self._interface.call_delete_subnet(net_index)

    async def set_key_phase(self, net_index: int, phase: int) -> None:
        await self._interface.call_set_key_phase(net_index, phase)

    async def create_app_key(self, net_index: int, app_index: int) -> None:
        await self._interface.call_create_app_key(net_index, app_index)

    async def import_app_key(
        self, net_index: int, app_index: int, app_key: ApplicationKey
    ) -> None:
        await self._interface.call_import_app_key(net_index, app_index, app_key.bytes)

    async def update_app_key(self, app_index: int) -> None:
        await self._interface.call_update_app_key(app_index)

    async def complete_app_key_update(self, app_index: int) -> None:
        await self._interface.call_comlete_app_key_update(app_index)

    async def delete_app_key(self, app_index: int) -> None:
        await self._interface.call_delete_app_key(app_index)

    async def import_remote_node(
        self, primary: int, count: int, device_key: DeviceKey
    ) -> None:
        await self._interface.call_import_remote_node(primary, count, device_key.bytes)

    async def delete_remote_node(self, primary: int, count: int) -> None:
        await self._interface.call_delete_remote_node(primary, count)


class TCPServerService:
    NAME = "org.bluez.mesh"

    def __init__(self, port):
        self.PATH = f"/org/bluez/mesh/tcpserver_{port}"


class AccessControlListInterface:
    def __init__(self, tcpserver_service):
        self._interface = tcpserver_service.get_interface(
            "org.bluez.mesh.AccessControlList1"
        )

    async def grant_access(
        self, uuid: UUID, dev_key: DeviceKey, net_key: NetworkKey
    ) -> int:
        return await self._interface.call_grant_access(
            uuid.bytes, dev_key.bytes, net_key.bytes
        )

    async def revoke_access(self, token: int) -> None:
        await self._interface.call_revoke_access(token)


class TCPConnectionService:
    NAME = "org.bluez.mesh"

    def __init__(self, tcpserver_service: TCPServerService, identity: str):
        self.PATH = f"{tcpserver_service.PATH}/{identity}"


class ConnectionStatInterface:
    def __init__(self, tcpconnection_service):
        self._interface = tcpconnection_service.get_interface(
            "org.bluez.mesh.ConnectionStat1"
        )

    async def connected(self) -> bool:
        return self._interface.connected

    async def last_error(self) -> str:
        return self._interface.last_error

    async def transmitted_msg_count(self) -> int:
        return self._interface.transmitted_msg_count

    async def received_msg_count(self) -> int:
        return self._interface.received_msg_count

    async def last_transmitted_msg_timestamp(self) -> int:
        return self._interface.last_transmitted_msg_timestamp

    async def last_received_msg_timestamp(self) -> int:
        return self._interface.last_received_msg_timestamp



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0621, C0103
from unittest import mock

import asynctest
from asynctest import ANY

from bluetooth_mesh import Element, SceneClient
from bluetooth_mesh.messages.config import GATTNamespaceDescriptor
from bluetooth_mesh.test.fixtures import *  # pylint: disable=W0614, W0401


class SceneElementMock(Element):
    LOCATION = GATTNamespaceDescriptor.BOTTOM
    MODELS = [SceneClient]


@pytest.fixture
def scene_client(element_path) -> SceneClient:
    element = SceneElementMock(mock.MagicMock(), mock.MagicMock())
    element.path = element_path
    return SceneClient(element)


@pytest.mark.asyncio
@asynctest.patch("asyncio.sleep", new=asynctest.CoroutineMock())
async def test_sending_scene_recall_repeated_6_times_with_intervals(
    scene_client, destination, app_index
):
    scene_client.repeat = asynctest.CoroutineMock()

    await scene_client.recall_scene_unack(
        destination=destination, app_index=app_index, scene_number=1, transition_time=0
    )

    scene_client.repeat.assert_awaited_once_with(
        ANY, retransmissions=6, send_interval=0.075
    )


# pylint: disable=W0212
@pytest.mark.asyncio
@asynctest.patch("asyncio.sleep", new=asynctest.CoroutineMock())
async def test_scene_recall_calls_node_interface_with_appropriate_arguments(
    scene_client, destination, app_index, element_path
):
    scene_client._node_interface.send = asynctest.CoroutineMock()

    await scene_client.recall_scene_unack(
        destination, app_index, scene_number=1, transition_time=0
    )

    data = b"\x82C\x01\x00\x00\x00\x18"
    scene_client._node_interface.send.assert_awaited_with(
        element_path, destination, app_index, data
    )


@pytest.mark.asyncio
@asynctest.patch("asyncio.sleep", new=asynctest.CoroutineMock())
@pytest.mark.parametrize(
    "number_of_scene_recalls, next_tid",
    [pytest.param(1, 0), pytest.param(255, 254), pytest.param(256, 0)],
)
async def test_scene_recall_increases_tid(
    scene_client, number_of_scene_recalls, next_tid
):
    scene_client.send_app = asynctest.CoroutineMock()

    for _ in range(number_of_scene_recalls):
        await scene_client.recall_scene_unack(
            ANY, ANY, scene_number=ANY, transition_time=ANY
        )

    scene_client.send_app.assert_awaited_with(
        ANY,
        app_index=ANY,
        opcode=ANY,
        params=dict(scene_number=ANY, tid=next_tid, transition_time=ANY, delay=ANY),
    )



from bluetooth_mesh.apps import hookimpl


class LocalNetworkMixin:
    async def get_network(self):
        # TODO: add a way to use a local provisioning database

        raise NotImplementedError(
            "Cannot obtain network database without Silvair commissioning API"
        )


@hookimpl(trylast=True)
def application_mixins():
    return (LocalNetworkMixin,)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Embedded, Int8ul, Int16ul, Select, Struct, Switch, this

from bluetooth_mesh.messages.config import StatusCodeAdapter
from bluetooth_mesh.messages.generics import (
    OptionalSetParameters,
    TransitionTime,
    TransitionTimeAdapter,
)
from bluetooth_mesh.messages.util import EnumAdapter, Opcode, OpcodeMessage


class LightLightnessOpcode(IntEnum):
    LIGHTNESS_GET = 0x824B
    LIGHTNESS_SET = 0x824C
    LIGHTNESS_SET_UNACKNOWLEDGED = 0x824D
    LIGHTNESS_STATUS = 0x824E
    LIGHTNESS_LINEAR_GET = 0x824F
    LIGHTNESS_LINEAR_SET = 0x8250
    LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED = 0x8251
    LIGHTNESS_LINEAR_STATUS = 0x8252
    LIGHTNESS_LAST_GET = 0x8253
    LIGHTNESS_LAST_STATUS = 0x8254
    LIGHTNESS_DEFAULT_GET = 0x8255
    LIGHTNESS_DEFAULT_STATUS = 0x8256
    LIGHTNESS_RANGE_GET = 0x8257
    LIGHTNESS_RANGE_STATUS = 0x8258

    def __repr__(self):
        return str(self.value)


class LightLightnessSetupOpcode(IntEnum):
    LIGHTNESS_DEFAULT_SET = 0x8259
    LIGHTNESS_DEFAULT_SET_UNACKNOWLEDGED = 0x825A
    LIGHTNESS_RANGE_SET = 0x825B
    LIGHTNESS_RANGE_SET_UNACKNOWLEDGED = 0x825C

    def __repr__(self):
        return str(self.value)


# fmt: off
LightLightnessGet = Struct()

LightLightnessDefault = Struct(
    "lightness" / Int16ul
)

LightLightnessStatusMinimal = Struct(
    "present_lightness" / Int16ul
)

LightLightnessStatusOptional = Struct(
    Embedded(LightLightnessStatusMinimal),
    "target_lightness" / Int16ul,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=True)
)

LightLightnessStatus = Select(
    LightLightnessStatusOptional,
    LightLightnessStatusMinimal
)

LightLightnessRange = Struct(
    "range_min" / Int16ul,
    "range_max" / Int16ul,
)

LightLightnessRangeStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(LightLightnessRange)
)

LightLightnessSetMinimal = Struct(
    "lightness" / Int16ul,
    "tid" / Int8ul
)

LightLightnessSetOptional = Struct(
    Embedded(LightLightnessSetMinimal),
    Embedded(OptionalSetParameters)
)

LightLightnessSet = Select(
    LightLightnessSetOptional,
    LightLightnessSetMinimal
)

LightLightnessMessage = OpcodeMessage({
    LightLightnessOpcode.LIGHTNESS_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHTNESS_SET: LightLightnessSet,
    LightLightnessOpcode.LIGHTNESS_SET_UNACKNOWLEDGED: LightLightnessSet,
    LightLightnessOpcode.LIGHTNESS_STATUS: LightLightnessStatus,
    LightLightnessOpcode.LIGHTNESS_LINEAR_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHTNESS_LINEAR_SET: LightLightnessSet,
    LightLightnessOpcode.LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED: LightLightnessSet,
    LightLightnessOpcode.LIGHTNESS_LINEAR_STATUS: LightLightnessStatus,
    LightLightnessOpcode.LIGHTNESS_LAST_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHTNESS_LAST_STATUS: LightLightnessDefault,
    LightLightnessOpcode.LIGHTNESS_DEFAULT_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHTNESS_DEFAULT_STATUS: LightLightnessDefault,
    LightLightnessOpcode.LIGHTNESS_RANGE_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHTNESS_RANGE_STATUS: LightLightnessRangeStatus,
})

LightLightnessSetupMessage = OpcodeMessage({
    LightLightnessSetupOpcode.LIGHTNESS_DEFAULT_SET: LightLightnessDefault,
    LightLightnessSetupOpcode.LIGHTNESS_DEFAULT_SET_UNACKNOWLEDGED: LightLightnessDefault,
    LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET: LightLightnessRange,
    LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET_UNACKNOWLEDGED: LightLightnessRange,
})
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.generic.onoff import GenericOnOffOpcode


@pytest.fixture()
def status_parsed():
    return dict(opcode=GenericOnOffOpcode.ONOFF_STATUS, params=dict(present_onoff=0))


@pytest.fixture()
def status_encoded():
    return b"\x82\x04\x00"


@pytest.fixture()
def get_encoded():
    return b"\x82\x01"


@pytest.fixture()
def source():
    return 0x0001


@pytest.fixture()
def destination():
    return 0x0010


@pytest.fixture()
def app_index():
    return 0


@pytest.fixture()
def net_index():
    return 0


@pytest.fixture
def element_path():
    return "test_element_path_string"



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from datetime import datetime

import pytest

from bluetooth_mesh.messages.properties import PropertyID
from bluetooth_mesh.messages.sensor import (
    SensorMessage,
    SensorOpcode,
    SensorSampling,
    SensorSettingAccess,
    SensorSetupMessage,
    SensorSetupOpcode,
)

valid = [
    # fmt: off
    pytest.param(
        b'\x82\x30',
        SensorOpcode.SENSOR_DESCRIPTOR_GET,
        dict(),
        id="SENSOR_DESCRIPTOR_GET"),
    pytest.param(
        b'\x82\x30\x04\x00',
        SensorOpcode.SENSOR_DESCRIPTOR_GET,
        dict(property_id=PropertyID.AVERAGE_OUTPUT_CURRENT),
        id="SENSOR_DESCRIPTOR_GET_with_optional"),
    pytest.param(
        b'\x82\x31',
        SensorOpcode.SENSOR_GET,
        dict(),
        id="SENSOR_GET"),
    pytest.param(
        b'\x82\x31\x07\x00',
        SensorOpcode.SENSOR_GET,
        dict(property_id=PropertyID.CHROMATICITY_TOLERANCE),
        id="SENSOR_GET_with_optional"),

    pytest.param(
        b'\x51\x0c\x00\x00\x00\x00\x04\x0b\x0c',
        SensorOpcode.SENSOR_DESCRIPTOR_STATUS,
        [dict(sensor_property_id=PropertyID.DEVICE_DATE_OF_MANUFACTURE,
              sensor_positive_tolerance=0x00,
              sensor_negative_tolerance=0x00,
              sensor_sampling_funcion=SensorSampling.MAXIMUM,
              sensor_measurement_period=0x0b,
              sensor_update_interval=0x0c)],
        id="SENSOR_DESCRIPTOR_STATUS"),
    pytest.param(
        b'\x51\x19\x00',
        SensorOpcode.SENSOR_DESCRIPTOR_STATUS,
        [dict(sensor_property_id=PropertyID.DEVICE_SERIAL_NUMBER)],
        id="SENSOR_DESCRIPTOR_STATUS_not_existing"),
    pytest.param(
        b'\x51\x0c\x00\x00\x00\x00\x02\x0b\x0c\x1f\x00\xef\xcd\xab\x07\x1b\x1c',
        SensorOpcode.SENSOR_DESCRIPTOR_STATUS,
        [dict(sensor_property_id=PropertyID.DEVICE_DATE_OF_MANUFACTURE,
              sensor_positive_tolerance=0x00,
              sensor_negative_tolerance=0x00,
              sensor_sampling_funcion=SensorSampling.ARITHMETIC_MEAN,
              sensor_measurement_period=0x0b,
              sensor_update_interval=0x0c),
         dict(sensor_property_id=PropertyID.INITIAL_LUMINOUS_FLUX,
              sensor_positive_tolerance=0xabc,
              sensor_negative_tolerance=0xdef,
              sensor_sampling_funcion=SensorSampling.COUNT,
              sensor_measurement_period=0x1b,
              sensor_update_interval=0x1c)],
        id="SENSOR_DESCRIPTOR_STATUS_multiple"),

    pytest.param(
        b'\x52\xe2\x0a\xc8\x00',
        SensorOpcode.SENSOR_STATUS,
        [dict(sensor_setting_property_id=PropertyID.PRESENT_INPUT_CURRENT,
              length=2,
              format=0,
              sensor_setting_raw=dict(
                  current=2.0))],
        id="SENSOR_STATUS_PRESENT_INPUT_CURRENT"),
    pytest.param(
        b'\x52\x22\x0b\x20\x03',
        SensorOpcode.SENSOR_STATUS,
        [dict(sensor_setting_property_id=PropertyID.PRESENT_INPUT_VOLTAGE,
              length=2,
              format=0,
              sensor_setting_raw=dict(
                  voltage=12.5))],
        id="SENSOR_STATUS_PRESENT_INPUT_VOLTAGE"),
    pytest.param(
        b'\x52\x44\x0d\xa2\x44\xff',
        SensorOpcode.SENSOR_STATUS,
        [dict(sensor_setting_property_id=PropertyID.TOTAL_DEVICE_ENERGY_USE,
              length=3,
              format=0,
              sensor_setting_raw=dict(
                  energy=0xff44a2))],
        id="SENSOR_STATUS_TOTAL_DEVICE_ENERGY_USE"),
    pytest.param(
        b'\x52\x09\x90\x40\xa2\x44\xff\x00\x00',
        SensorOpcode.SENSOR_STATUS,
        [dict(sensor_setting_property_id=0x4090,
              length=5,
              format=1,
              sensor_setting_raw=list(b'\xa2\x44\xff\x00\x00'))],
        id="SENSOR_STATUS_VENDOR_PROPERTY"),
    pytest.param(
        b'\x52\x44\x0d\xa2\x44\xff\x22\x0b\x20\x03',
        SensorOpcode.SENSOR_STATUS,
        [dict(sensor_setting_property_id=PropertyID.TOTAL_DEVICE_ENERGY_USE,
              length=3,
              format=0,
              sensor_setting_raw=dict(
                  energy=0xff44a2)),
         dict(sensor_setting_property_id=PropertyID.PRESENT_INPUT_VOLTAGE,
              length=2,
              format=0,
              sensor_setting_raw=dict(
                  voltage=12.5))
         ],
        id="SENSOR_STATUS_2_SHORT_PROP"),
    # fmt: on
]

valid_setup = [
    # fmt: off
    pytest.param(
        b'\x58\x30\x00\x01\x00\x04\x00\x09\x00',
        SensorSetupOpcode.SENSOR_SETTINGS_STATUS,
        dict(sensor_property_id=PropertyID.LIGHT_CONTROL_LIGHTNESS_STANDBY,
             sensor_setting_property_ids=[PropertyID.AVERAGE_AMBIENT_TEMPERATURE_IN_A_PERIOD_OF_DAY,
                                          PropertyID.AVERAGE_OUTPUT_CURRENT,
                                          PropertyID.COLOR_RENDERING_INDEX_RA]),
        id="SENSOR_SETTINGS_STATUS"),
    pytest.param(
        b'\x58\x30\x00',
        SensorSetupOpcode.SENSOR_SETTINGS_STATUS,
        dict(sensor_property_id=PropertyID.LIGHT_CONTROL_LIGHTNESS_STANDBY,
             sensor_setting_property_ids=[]),
        id="SENSOR_SETTING_SET_empty"),
    pytest.param(
        b'\x59\x57\x00\x57\x00\xc8\x00',
        SensorSetupOpcode.SENSOR_SETTING_SET,
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_raw=dict(
                 current=2.0)),
        id="SENSOR_SETTING_SET"),
    pytest.param(
        b'\x5b\x57\x00\x57\x00\x01\xc8\x00',
        SensorSetupOpcode.SENSOR_SETTING_STATUS,
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_access=SensorSettingAccess.READ_ONLY,
             sensor_setting_raw=dict(
                 current=2.0)),
        id="SENSOR_SETTING_STATUS"),
    pytest.param(
        b'\x5b\x57\x00\x02\x00\x01\xc8\x00\x39',
        SensorSetupOpcode.SENSOR_SETTING_STATUS,
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_property_id=PropertyID.AVERAGE_INPUT_CURRENT,
             sensor_setting_access=SensorSettingAccess.READ_ONLY,
             sensor_setting_raw=dict(
                 electric_current_value=2.0,
                 sensing_duration=dict(seconds=0.5132))),
        id="SENSOR_SETTING_STATUS"),
    # fmt: on
]

valid_properties = [
    # fmt: off
    pytest.param(
        b'\x59\x57\x00\x57\x00\xc8\x00',
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_raw=dict(
                 current=2.0)),
        id="ElectricCurrent"),
    pytest.param(
        b'\x59\x57\x00\x02\x00\xc8\x00\x39',
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_property_id=PropertyID.AVERAGE_INPUT_CURRENT,
             sensor_setting_raw=dict(
                 electric_current_value=2.0,
                 sensing_duration=dict(seconds=0.5132))),
        id="AverageCurrent"),
    pytest.param(
        b'\x59\x05\x00\x05\x00\x20\x03\x54',
        dict(sensor_property_id=PropertyID.AVERAGE_OUTPUT_VOLTAGE,
             sensor_setting_property_id=PropertyID.AVERAGE_OUTPUT_VOLTAGE,
             sensor_setting_raw=dict(
                 voltage_value=12.5,
                 sensing_duration=dict(seconds=6.7275))),
        id="AverageVoltage"),
    pytest.param(
        b'\x59\x57\x00\x02\x00\xc8\x00\x00',
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_CURRENT,
             sensor_setting_property_id=PropertyID.AVERAGE_INPUT_CURRENT,
             sensor_setting_raw=dict(
                 electric_current_value=2.0,
                 sensing_duration=dict(seconds=0))),
        id="AverageCurrent_with_0"),
    pytest.param(
        b'\x59\x59\x00\x59\x00\x00\x03',
        dict(sensor_property_id=PropertyID.PRESENT_INPUT_VOLTAGE,
             sensor_setting_property_id=PropertyID.PRESENT_INPUT_VOLTAGE,
             sensor_setting_raw=dict(
                 voltage=12)),
        id="Voltage"),
    pytest.param(
        b'\x59\x42\x00\x4d\x00\x01',
        dict(sensor_property_id=PropertyID.MOTION_SENSED,
             sensor_setting_property_id=PropertyID.PRESENCE_DETECTED,
             sensor_setting_raw=dict(
                 presence_detected=True)),
        id="Presence"),
    pytest.param(
        b'\x59\x42\x00\x42\x00\x50',
        dict(sensor_property_id=PropertyID.MOTION_SENSED,
             sensor_setting_property_id=PropertyID.MOTION_SENSED,
             sensor_setting_raw=dict(
                 percentage=40)),
        id="Percentage8"),
    pytest.param(
        b'\x59\x0a\x00\x36\x00\xb8\x0b\x00',
        dict(sensor_property_id=PropertyID.DEVICE_APPEARANCE,
             sensor_setting_property_id=PropertyID.LIGHT_CONTROL_TIME_FADE,
             sensor_setting_raw=dict(
                 seconds=3)),
        id="TimeMiliseconds24"),
    pytest.param(
        b'\x59\x6d\x00\x6d\x00\x0a\x00\x00',
        dict(sensor_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_TIME,
             sensor_setting_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_TIME,
             sensor_setting_raw=dict(
                 hours=10)),
        id="TimeHour24"),
    pytest.param(
        b'\x59\x6d\x00\x6d\x00\xff\xff\xff',
        dict(sensor_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_TIME,
             sensor_setting_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_TIME,
             sensor_setting_raw=dict(
                 hours=None)),
        id="TimeHour24_unknown"),
    pytest.param(
        b'\x59\x55\x00\x55\x00\x1a\x27\x00',
        dict(sensor_property_id=PropertyID.PRESENT_ILLUMINANCE,
             sensor_setting_property_id=PropertyID.PRESENT_ILLUMINANCE,
             sensor_setting_raw=dict(
                 illuminance=100.1)),
        id="Illuminance"),
    pytest.param(
        b'\x59\x4c\x00\x4c\x00\x1b\x1a',
        dict(sensor_property_id=PropertyID.PEOPLE_COUNT,
             sensor_setting_property_id=PropertyID.PEOPLE_COUNT,
             sensor_setting_raw=dict(
                 count=0x1a1b)),
        id="Count16"),
    pytest.param(
        b'\x59\x6c\x00\x6c\x00\xff\x1b\x1a',
        dict(sensor_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_CYCLES,
             sensor_setting_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_CYCLES,
             sensor_setting_raw=dict(
                 count=0x1a1bff)),
        id="Count24"),
    pytest.param(
        b'\x59\x6c\x00\x6c\x00\xff\xff\xff',
        dict(sensor_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_CYCLES,
             sensor_setting_property_id=PropertyID.TOTAL_DEVICE_POWER_ON_CYCLES,
             sensor_setting_raw=dict(
                 count=None)),
        id="Count24_unknown"),
    pytest.param(
        b'\x59\x68\x00\x68\x00\x05\x00',
        dict(sensor_property_id=PropertyID.TIME_SINCE_MOTION_SENSED,
             sensor_setting_property_id=PropertyID.TIME_SINCE_MOTION_SENSED,
             sensor_setting_raw=dict(
                 seconds=5)),
        id="TimeSecond16"),
    pytest.param(
        b'\x59\x67\x00\x67\x00\x05\x00\x01\x00\x70\x40',
        dict(sensor_property_id=PropertyID.SHORT_CIRCUIT_EVENT_STATISTICS,
             sensor_setting_property_id=PropertyID.SHORT_CIRCUIT_EVENT_STATISTICS,
             sensor_setting_raw=dict(
                 number_of_events=dict(count=5),
                 average_event_duration=dict(seconds=1),
                 time_elapsed_since_last_event=dict(seconds=97.0172),
                 sensing_duration=dict(seconds=1),
             )),
        id="EventStatistics"),
    pytest.param(
        b'\x59\x0e\x00\x0e\x00abcdefgh',
        dict(sensor_property_id=PropertyID.DEVICE_FIRMWARE_REVISION,
             sensor_setting_property_id=PropertyID.DEVICE_FIRMWARE_REVISION,
             sensor_setting_raw='abcdefgh',
             ),
        id="FixedString8"),
    pytest.param(
        b'\x59\x0e\x00\x0e\x00abcdef\x00\x00',
        dict(sensor_property_id=PropertyID.DEVICE_FIRMWARE_REVISION,
             sensor_setting_property_id=PropertyID.DEVICE_FIRMWARE_REVISION,
             sensor_setting_raw='abcdef',
             ),
        id="FixedString8_padded"),
    pytest.param(
        b'\x59\x11\x00\x11\x00abcdefghabcdefghabcdefghabcdefghabcd',
        dict(sensor_property_id=PropertyID.DEVICE_MANUFACTURER_NAME,
             sensor_setting_property_id=PropertyID.DEVICE_MANUFACTURER_NAME,
             sensor_setting_raw='abcdefghabcdefghabcdefghabcdefghabcd',
             ),
        id="FixedString36"),
    pytest.param(
        b'\x59\x6a\x00\x6a\x00\xa2\x44\xff',
        dict(sensor_property_id=PropertyID.TOTAL_DEVICE_ENERGY_USE,
             sensor_setting_property_id=PropertyID.TOTAL_DEVICE_ENERGY_USE,
             sensor_setting_raw=dict(
                 energy=0xff44a2
             )),
        id="Energy"),
    pytest.param(
        b'\x59\x2e\x00\x2e\x00\x44\xff',
        dict(sensor_property_id=PropertyID.LIGHT_CONTROL_LIGHTNESS_ON,
             sensor_setting_property_id=PropertyID.LIGHT_CONTROL_LIGHTNESS_ON,
             sensor_setting_raw=dict(
                 perceived_lightness=0xff44
             )),
        id="PerceivedLightness"),
    pytest.param(
        b'\x59\x32\x00\x32\x00\x44\xff\x00\x00',
        dict(sensor_property_id=PropertyID.LIGHT_CONTROL_REGULATOR_KID,
             sensor_setting_property_id=PropertyID.LIGHT_CONTROL_REGULATOR_KID,
             sensor_setting_raw=dict(
                 coefficient=0x0000ff44
             )),
        id="Coefficient"),
    pytest.param(
        b'\x59\x52\x00\x52\x00\xa0\x86\x01',
        dict(sensor_property_id=PropertyID.PRESENT_DEVICE_INPUT_POWER,
             sensor_setting_property_id=PropertyID.PRESENT_DEVICE_INPUT_POWER,
             sensor_setting_raw=dict(
                 power=10000
             )),
        id="Power"),
    pytest.param(
        b'\x59\x16\x00\x16\x00\xe8\x03\x00\xd0\x07\x00\x0f\x27\x00',
        dict(sensor_property_id=PropertyID.DEVICE_POWER_RANGE_SPECIFICATION,
             sensor_setting_property_id=PropertyID.DEVICE_POWER_RANGE_SPECIFICATION,
             sensor_setting_raw=dict(
                 minimum_power_value=100,
                 typical_power_value=200,
                 maximum_power_value=999.9,
             )),
        id="PowerSpecification"),
    pytest.param(
        b'\x59\x4f\x00\x4f\x00\x1f',
        dict(sensor_property_id=PropertyID.PRESENT_AMBIENT_TEMPERATURE,
             sensor_setting_property_id=PropertyID.PRESENT_AMBIENT_TEMPERATURE,
             sensor_setting_raw=dict(
                 temperature=15.5,
             )),
        id="Temperature8_positive"),
    pytest.param(
        b'\x59\x4f\x00\x4f\x00\xe1',
        dict(sensor_property_id=PropertyID.PRESENT_AMBIENT_TEMPERATURE,
             sensor_setting_property_id=PropertyID.PRESENT_AMBIENT_TEMPERATURE,
             sensor_setting_raw=dict(
                 temperature=-15.5,
             )),
        id="Temperature8_negative"),
    pytest.param(
        b'\x59\x54\x00\x54\x00\xe6\x20',
        dict(sensor_property_id=PropertyID.PRESENT_DEVICE_OPERATING_TEMPERATURE,
             sensor_setting_property_id=PropertyID.PRESENT_DEVICE_OPERATING_TEMPERATURE,
             sensor_setting_raw=dict(
                 temperature=84.22,
             )),
        id="Temperature_positive"),
    pytest.param(
        b'\x59\x45\x00\x45\x00\x31\x02\xdc\x6e\x71',
        dict(sensor_property_id=PropertyID.OUTDOOR_STATISTICAL_VALUES,
             sensor_setting_property_id=PropertyID.OUTDOOR_STATISTICAL_VALUES,
             sensor_setting_raw=dict(
                 average_temperature=24.5,
                 standard_deviation_temperature=1,
                 minimum_temperature=-18,
                 maximum_temperature=55,
                 sensing_duration=dict(seconds=106.719),
             )),
        id="Temperature8Statistics"),
    pytest.param(
        b'\x59\x14\x00\x14\x00\x92\x09\x64\x00\xf8\xf8\x20\x4e\x71',
        dict(sensor_property_id=PropertyID.DEVICE_OPERATING_TEMPERATURE_STATISTICAL_VALUES,
             sensor_setting_property_id=PropertyID.DEVICE_OPERATING_TEMPERATURE_STATISTICAL_VALUES,
             sensor_setting_raw=dict(
                 average_temperature=24.5,
                 standard_deviation_temperature=1,
                 minimum_temperature=-18,
                 maximum_temperature=200,
                 sensing_duration=dict(seconds=106.719),
             )),
        id="TemperatureStatistics"),
    pytest.param(
        b'\x59\x13\x00\x13\x00\xdc\x6e',
        dict(sensor_property_id=PropertyID.DEVICE_OPERATING_TEMPERATURE_RANGE_SPECIFICATION,
             sensor_setting_property_id=PropertyID.DEVICE_OPERATING_TEMPERATURE_RANGE_SPECIFICATION,
             sensor_setting_raw=dict(
                 minimum_temperature=-18,
                 maximum_temperature=55,
             )),
        id="TemperatureRange"),
    pytest.param(
        b'\x59\x01\x00\x01\x00\x18\x6d\xa2',
        dict(sensor_property_id=PropertyID.AVERAGE_AMBIENT_TEMPERATURE_IN_A_PERIOD_OF_DAY,
             sensor_setting_property_id=PropertyID.AVERAGE_AMBIENT_TEMPERATURE_IN_A_PERIOD_OF_DAY,
             sensor_setting_raw=dict(
                 temperature=12,
                 start_time=dict(hour=10.9),
                 end_time=dict(hour=16.2),
             )),
        id="Temperature8InAPeriodOfDay"),
    pytest.param(
        b'\x59\x60\x00\x60\x00\xa2\x44\xff\x6d\xa2',
        dict(sensor_property_id=PropertyID.RELATIVE_DEVICE_ENERGY_USE_IN_A_PERIOD_OF_DAY,
             sensor_setting_property_id=PropertyID.RELATIVE_DEVICE_ENERGY_USE_IN_A_PERIOD_OF_DAY,
             sensor_setting_raw=dict(
                 energy_value=0xff44a2,
                 start_time=dict(hour=10.9),
                 end_time=dict(hour=16.2),
             )),
        id="EnergyInAPeriodOfDay"),
    pytest.param(
        b'\x59\x2a\x00\x2a\x00\x69\x01\x40\x00\x00\x00\xf0\xff\x54',
        dict(sensor_property_id=PropertyID.INPUT_VOLTAGE_STATISTICS,
             sensor_setting_property_id=PropertyID.INPUT_VOLTAGE_STATISTICS,
             sensor_setting_raw=dict(
                 average_voltage_value=5.640625,
                 standard_deviation_voltage_value=1,
                 minimum_voltage_value=0,
                 maximum_voltage_value=1023.75,
                 sensing_duration=dict(seconds=6.7275),
             )),
        id="VoltageStatistics"),
    pytest.param(
        b'\x59\x49\x00\x49\x00\x00\x00\x00\x06\xf0\xff',
        dict(sensor_property_id=PropertyID.OUTPUT_VOLTAGE_RANGE,
             sensor_setting_property_id=PropertyID.OUTPUT_VOLTAGE_RANGE,
             sensor_setting_raw=dict(
                 minimum_voltage_value=0,
                 typical_voltage_value=24,
                 maximum_voltage_value=1023.75,
             )),
        id="VoltageRange"),
    pytest.param(
        b'\x59\x47\x00\x47\x00\x97\x09\x64\x00\x00\x00\x7d\x15\x71',
        dict(sensor_property_id=PropertyID.OUTPUT_CURRENT_STATISTICS,
             sensor_setting_property_id=PropertyID.OUTPUT_CURRENT_STATISTICS,
             sensor_setting_raw=dict(
                 average_electric_current_value=24.55,
                 standard_deviation_electric_current_value=1,
                 minimum_electric_current_value=0,
                 maximum_electric_current_value=55.01,
                 sensing_duration=dict(seconds=106.719),
             )),
        id="ElectricCurrentStatistics"),
    pytest.param(
        b'\x59\x46\x00\x46\x00\x00\x00\x7d\x15',
        dict(sensor_property_id=PropertyID.OUTPUT_CURRENT_RANGE,
             sensor_setting_property_id=PropertyID.OUTPUT_CURRENT_RANGE,
             sensor_setting_raw=dict(
                 minimum_electric_current_value=0,
                 maximum_electric_current_value=55.01,
             )),
        id="ElectricCurrentRange"),
    pytest.param(
        b'\x59\x21\x00\x21\x00\x00\x00\x01\x00\x7d\x15',
        dict(sensor_property_id=PropertyID.INPUT_CURRENT_RANGE_SPECIFICATION,
             sensor_setting_property_id=PropertyID.INPUT_CURRENT_RANGE_SPECIFICATION,
             sensor_setting_raw=dict(
                 minimum_electric_current_value=0,
                 typical_electric_current_value=0.01,
                 maximum_electric_current_value=55.01,
             )),
        id="ElectricCurrentSpecification"),
    pytest.param(
        b'\x59\x70\x00\x70\x00\x02\x00\x00',
        dict(sensor_property_id=PropertyID.TOTAL_LUMINOUS_ENERGY,
             sensor_setting_property_id=PropertyID.TOTAL_LUMINOUS_ENERGY,
             sensor_setting_raw=dict(
                 luminous_energy=2000,
             )),
        id="LuminousEnergy"),
    pytest.param(
        b'\x59\x06\x00\x06\x00\x00\x20',
        dict(sensor_property_id=PropertyID.CENTER_BEAM_INTENSITY_AT_FULL_POWER,
             sensor_setting_property_id=PropertyID.CENTER_BEAM_INTENSITY_AT_FULL_POWER,
             sensor_setting_raw=dict(
                 luminous_intensity=0x2000,
             )),
        id="LuminousIntensity"),
    pytest.param(
        b'\x59\x40\x00\x40\x00\x02\x00\x00',
        dict(sensor_property_id=PropertyID.LUMINOUS_EXPOSURE,
             sensor_setting_property_id=PropertyID.LUMINOUS_EXPOSURE,
             sensor_setting_raw=dict(
                 luminous_exposure=2000,
             )),
        id="LuminousExposure"),
    pytest.param(
        b'\x59\x1f\x00\x1f\x00\xd0\x07',
        dict(sensor_property_id=PropertyID.INITIAL_LUMINOUS_FLUX,
             sensor_setting_property_id=PropertyID.INITIAL_LUMINOUS_FLUX,
             sensor_setting_raw=dict(
                 luminous_flux=2000,
             )),
        id="LuminousFlux"),
    pytest.param(
        b'\x59\x41\x00\x41\x00\xe8\x03\xd0\x07',
        dict(sensor_property_id=PropertyID.LUMINOUS_FLUX_RANGE,
             sensor_setting_property_id=PropertyID.LUMINOUS_FLUX_RANGE,
             sensor_setting_raw=dict(
                 minimum_luminous_flux=1000,
                 maximum_luminous_flux=2000,
             )),
        id="LuminousFluxRange"),
    pytest.param(
        b'\x59\x3e\x00\x3e\x00\xd4\x07',
        dict(sensor_property_id=PropertyID.LUMINOUS_EFFICACY,
             sensor_setting_property_id=PropertyID.LUMINOUS_EFFICACY,
             sensor_setting_raw=dict(
                 luminous_efficacy=200.4,
             )),
        id="LuminousEfficacy"),
    pytest.param(
        b'\x59\x0f\x00\x0f\x00\xff\xee\xdd\xcc\xbb\xaa',
        dict(sensor_property_id=PropertyID.DEVICE_GLOBAL_TRADE_ITEM_NUMBER,
             sensor_setting_property_id=PropertyID.DEVICE_GLOBAL_TRADE_ITEM_NUMBER,
             sensor_setting_raw=dict(
                 global_trade_item_number=0xaabbccddeeff
             )),
        id="GlobalTradeItemNumber"),
    pytest.param(
        b'\x59\x07\x00\x07\x00\x64',
        dict(sensor_property_id=PropertyID.CHROMATICITY_TOLERANCE,
             sensor_setting_property_id=PropertyID.CHROMATICITY_TOLERANCE,
             sensor_setting_raw=dict(
                 chromaticity_tolerance=0.01,
             )),
        id="ChromaticityTolerance"),
    pytest.param(
        b'\x59\x5e\x00\x5e\x00\x92\x27',
        dict(sensor_property_id=PropertyID.PRESENT_PLANCKIAN_DISTANCE,
             sensor_setting_property_id=PropertyID.PRESENT_PLANCKIAN_DISTANCE,
             sensor_setting_raw=dict(
                 distance_from_planckian=0.1013,
             )),
        id="ChromaticDistanceFromPlanckian"),
    pytest.param(
        b'\x59\x51\x00\x51\x00\xb8\x0b',
        dict(sensor_property_id=PropertyID.PRESENT_CORRELATED_COLOR_TEMPERATURE,
             sensor_setting_property_id=PropertyID.PRESENT_CORRELATED_COLOR_TEMPERATURE,
             sensor_setting_raw=dict(
                 correlated_color_temperature=3000,
             )),
        id="CorrelatedColorTemperature"),
    pytest.param(
        b'\x59\x0a\x00\x0a\x00\x04\xf0',
        dict(sensor_property_id=PropertyID.DEVICE_APPEARANCE,
             sensor_setting_property_id=PropertyID.DEVICE_APPEARANCE,
             sensor_setting_raw=dict(
                 category=960,
                 sub_category=4
             )),
        id="Appearance"),
    pytest.param(
        b'\x59\x0b\x00\x0b\x00\x2a\x00',
        dict(sensor_property_id=PropertyID.DEVICE_COUNTRY_OF_ORIGIN,
             sensor_setting_property_id=PropertyID.DEVICE_COUNTRY_OF_ORIGIN,
             sensor_setting_raw=dict(
                 country_code=42,
             )),
        id="CountryCode"),
    pytest.param(
        b'\x59\x0c\x00\x0c\x00\xde\x46\x00',
        dict(sensor_property_id=PropertyID.DEVICE_DATE_OF_MANUFACTURE,
             sensor_setting_property_id=PropertyID.DEVICE_DATE_OF_MANUFACTURE,
             sensor_setting_raw=dict(
                 date=datetime(2019, 9, 3),
             )),
        id="DateUTC"),
    pytest.param(
        b'\x59\x50\x00\x50\x00\xee\x00\xcd\xab',
        dict(sensor_property_id=PropertyID.PRESENT_CIE_1931_CHROMATICITY_COORDINATES,
             sensor_setting_property_id=PropertyID.PRESENT_CIE_1931_CHROMATICITY_COORDINATES,
             sensor_setting_raw=dict(
                 chromaticity_x_coordinate=0xee/0xffff,
                 chromaticity_y_coordinate=0xabcd/0xffff,
             )),
        id="ChromaticityCoordinates"),
    pytest.param(
        b'\x59\x08\x00\x08\x00\x64',
        dict(sensor_property_id=PropertyID.COLOR_RENDERING_INDEX_R9,
             sensor_setting_property_id=PropertyID.COLOR_RENDERING_INDEX_R9,
             sensor_setting_raw=dict(
                 color_rendering_index=100,
             )),
        id="ColorRenderingIndex_positive"),
    pytest.param(
        b'\x59\x08\x00\x08\x00\x9c',
        dict(sensor_property_id=PropertyID.COLOR_RENDERING_INDEX_R9,
             sensor_setting_property_id=PropertyID.COLOR_RENDERING_INDEX_R9,
             sensor_setting_raw=dict(
                 color_rendering_index=-100,
             )),
        id="ColorRenderingIndex_negative"),
    pytest.param(
        b'\x59\x61\x00\x61\x00\x88\xaa\x00\xbb\xbb',
        dict(sensor_property_id=PropertyID.RELATIVE_DEVICE_RUNTIME_IN_A_GENERIC_LEVEL_RANGE,
             sensor_setting_property_id=PropertyID.RELATIVE_DEVICE_RUNTIME_IN_A_GENERIC_LEVEL_RANGE,
             sensor_setting_raw=dict(
                 relative_value=0x44,
                 minimum_generic_level=0xaa,
                 maximum_generic_level=0xbbbb,
             )),
        id="RelativeRuntimeInAGenericLevelRange"),
    pytest.param(
        b'\x59\x62\x00\x62\x00\x88\x1a\x27\x00\x1a\x27\x00',
        dict(sensor_property_id=PropertyID.RELATIVE_EXPOSURE_TIME_IN_AN_ILLUMINANCE_RANGE,
             sensor_setting_property_id=PropertyID.RELATIVE_EXPOSURE_TIME_IN_AN_ILLUMINANCE_RANGE,
             sensor_setting_raw=dict(
                 relative_value=0x44,
                 minimum_illuminance=100.1,
                 maximum_illuminance=100.1,
             )),
        id="RelativeValueInAnIlluminanceRange"),
    pytest.param(
        b'\x59\x64\x00\x64\x00\x88\xff\xff\xe6\x20',
        dict(sensor_property_id=PropertyID.RELATIVE_RUNTIME_IN_A_DEVICE_OPERATING_TEMPERATURE_RANGE,
             sensor_setting_property_id=PropertyID.RELATIVE_RUNTIME_IN_A_DEVICE_OPERATING_TEMPERATURE_RANGE,
             sensor_setting_raw=dict(
                 relative_value=0x44,
                 minimum_temperature=-0.01,
                 maximum_temperature=84.22,
             )),
        id="RelativeValueInATemperatureRange"),
    pytest.param(
        b'\x59\x65\x00\x65\x00\x88\x00\x00\xcd\xab',
        dict(sensor_property_id=PropertyID.RELATIVE_RUNTIME_IN_AN_INPUT_CURRENT_RANGE,
             sensor_setting_property_id=PropertyID.RELATIVE_RUNTIME_IN_AN_INPUT_CURRENT_RANGE,
             sensor_setting_raw=dict(
                 relative_value=0x44,
                 minimum_current=0,
                 maximum_current=439.81,
             )),
        id="RelativeValueInACurrentRange"),
    pytest.param(
        b'\x59\x66\x00\x66\x00\x88\x20\x03\x20\x03',
        dict(sensor_property_id=PropertyID.RELATIVE_RUNTIME_IN_AN_INPUT_VOLTAGE_RANGE,
             sensor_setting_property_id=PropertyID.RELATIVE_RUNTIME_IN_AN_INPUT_VOLTAGE_RANGE,
             sensor_setting_raw=dict(
                 relative_value=0x44,
                 minimum_voltage=12.5,
                 maximum_voltage=12.5
             )),
        id="RelativeValueInAVoltageRange"),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert SensorMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert SensorMessage.build(dict(opcode=opcode, params=data)) == encoded


@pytest.mark.parametrize("encoded,opcode,data", valid_setup)
def test_parse_valid_setup(encoded, opcode, data):
    assert SensorSetupMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid_setup)
def test_build_valid_setup(encoded, opcode, data):
    assert SensorSetupMessage.build(dict(opcode=opcode, params=data)) == encoded


@pytest.mark.parametrize("encoded,data", valid_properties)
def test_parse_valid_property(encoded, data):
    assert SensorSetupMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,data", valid_properties)
def test_build_valid_property(encoded, data):
    assert (
        SensorSetupMessage.build(
            dict(opcode=SensorSetupOpcode.SENSOR_SETTING_SET, params=data)
        )
        == encoded
    )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import asyncio
import itertools
import logging
from inspect import isawaitable
from typing import Any, Callable, Hashable, Iterable, Mapping, Optional, TypeVar

ParsedMeshMessage = Mapping[str, Any]
MessageDescription = Mapping[str, Any]
ProgressCallback = Callable[[Hashable, Any, Mapping, Mapping], None]
T = TypeVar("T")


def chunks(iterable: Iterable[T], n: int, fillvalue: Optional[T] = None):
    """Collect data into fixed-length chunks or blocks"""
    # chunks('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


class MeshError(Exception):
    pass


class ModelOperationError(MeshError):
    def __init__(self, message: str, status: Optional[ParsedMeshMessage] = None):
        super().__init__(message)
        self.status = status

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.status)


class Signal:
    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)
        self.callbacks = set()

    def connect(self, callback):
        self.callbacks.add(callback)

    def disconnect(self, callback):
        try:
            self.callbacks.remove(callback)
        except KeyError:
            pass

    async def emit(self, *args, **kwargs):
        for callback in list(self.callbacks):
            try:
                result = callback(*args, **kwargs)
                if isawaitable(result):
                    await result
            except Exception as ex:
                self.logger.warning("Callback exception: %s", ex)

    @property
    def connections(self):
        return len(self.callbacks)


class Gatherer:
    """
        Gather tasks as they are done and return them in asynchronous iterator form.

        :param aws: iterable collection of tasks
        :param timeout: Timeout in seconds of None for no timeout
        :param loop: asyncio loop or None

        :return: A tuple of:
            - task to collect
            - result of the task
    """

    def __init__(self, aws: Iterable, timeout: int = None, loop=None):
        self.tasks = {asyncio.ensure_future(task): task for task in aws}
        self.buffer = asyncio.Queue()

        for status in self.tasks.keys():
            status.add_done_callback(self._done_callback)

        if timeout:
            loop = loop or asyncio.get_event_loop()
            loop.call_later(timeout, self.timeout)

    def _done_callback(self, item):
        task = self.tasks.pop(item)
        try:
            self.buffer.put_nowait((task, item.exception() or item.result()))
        except asyncio.CancelledError:
            self.buffer.put_nowait((task, TimeoutError()))

    def timeout(self):
        for task in self.tasks.keys():
            if not task.done():
                task.cancel()

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.buffer.qsize() and not self.tasks:
            raise StopAsyncIteration

        return await self.buffer.get()


def construct_match(received, expected):
    def match(lhs, rhs):

        if rhs is Any or rhs is ...:
            return True

        if isinstance(rhs, list):
            return len(rhs) == len(lhs) and all(match(*i) for i in zip(lhs, rhs))

        if isinstance(rhs, dict):
            _lhs = {k: v for k, v in lhs.items() if not k.startswith("_")}
            return all(match(v, rhs.get(k, Any)) for k, v in _lhs.items())

        return lhs == rhs

    return match(received, expected)



# pylint: disable=W0611

from .ctl import (
    LightCTLMessage,
    LightCTLOpcode,
    LightCTLSetupMessage,
    LightCTLSetupOpcode,
)
from .lightness import (
    LightLightnessMessage,
    LightLightnessOpcode,
    LightLightnessSetupMessage,
    LightLightnessSetupOpcode,
)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.generic.battery import *

valid = [
    # fmt: off
    pytest.param(
        b'\x82\x23',
        GenericBatteryOpcode.BATTERY_GET,
        dict(),
        id="BATTERY_GET"
    ),
    pytest.param(
        b'\x82\x24\x32\xb4\x00\x00\xfe\xfe\x00\x62',
        GenericBatteryOpcode.BATTERY_STATUS,
        dict(
            battery_level=50,
            time_to_discharge=0xb4,
            time_to_charge=0xfefe,
            flags=dict(
                battery_presence_flags=GenericBatteryFlagsPresence.BATTERY_PRESENT_NON_REMOVABLE,
                battery_indicator_flags=GenericBatteryFlagsIndicator.BATTERY_CHARGE_CRITICALLY_LOW,
                battery_charging_flags=GenericBatteryFlagsCharging.BATTERY_CHARGEABLE_CHARGING,
                battery_serviceability_flags=GenericBatteryFlagsServiceability.BATTERY_NOT_REQUIRE_SERVICE
            )
        ),
        id="BATTERY_STATUS"
    ),
    pytest.param(
        b'\x82\x24\xff\xbb\xaa\x00\xff\xff\xff\xdb',
        GenericBatteryOpcode.BATTERY_STATUS,
        dict(
            battery_level=None,
            time_to_discharge=0xaabb,
            time_to_charge=None,
            flags=dict(
                battery_presence_flags=GenericBatteryFlagsPresence.BATTERY_PRESENCE_UNKNOWN,
                battery_indicator_flags=GenericBatteryFlagsIndicator.BATTERY_CHARGE_GOOD,
                battery_charging_flags=GenericBatteryFlagsCharging.BATTERY_CHARGEABLE_NOT_CHARGING,
                battery_serviceability_flags=GenericBatteryFlagsServiceability.BATTERY_SERVICEABILITY_UNKNOWN
            )
        ),
        id="BATTERY_STATUS"
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert GenericBatteryMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert GenericBatteryMessage.build(dict(opcode=opcode, params=data)) == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
"""
This module implements mesh models, both clients and servers.
"""
import inspect
import itertools
from functools import partial
from typing import Any, Dict, Iterable, NamedTuple, Optional, Sequence, Tuple, Type

from construct import BitStruct

from bluetooth_mesh.crypto import ApplicationKey, NetworkKey
from bluetooth_mesh.messages.config import (
    CompositionData,
    ConfigOpcode,
    PublishFriendshipCredentialsFlag,
    SecureNetworkBeacon,
    StatusCode,
)
from bluetooth_mesh.messages.generic.level import GenericLevelOpcode
from bluetooth_mesh.messages.generic.light.ctl import LightCTLOpcode
from bluetooth_mesh.messages.generic.light.lightness import LightLightnessOpcode, LightLightnessSetupOpcode
from bluetooth_mesh.messages.generic.onoff import GenericOnOffOpcode
from bluetooth_mesh.messages.health import HealthOpcode
from bluetooth_mesh.messages.scene import SceneOpcode
from bluetooth_mesh.messages.sensor import SensorOpcode
from bluetooth_mesh.messages.silvair.debug import DebugOpcode, DebugSubOpcode
from bluetooth_mesh.messages.silvair.gateway_config_server import (
    DhcpFlag,
    GatewayConfigServerOpcode,
    GatewayConfigServerSubOpcode,
)
from bluetooth_mesh.messages.silvair.network_diagnostic_server import (
    NetworkDiagnosticServerOpcode,
    NetworkDiagnosticSetupServerOpcode,
)
from bluetooth_mesh.models.base import Model
from bluetooth_mesh.utils import ModelOperationError, ProgressCallback

__all__ = [
    "ConfigServer",
    "ConfigClient",
    "HealthServer",
    "HealthClient",
    "DebugServer",
    "DebugClient",
    "NetworkDiagnosticClient",
    "NetworkDiagnosticSetupClient",
    "GenericOnOffServer",
    "GenericOnOffClient",
    "SceneClient",
    "GenericLevelClient",
    "SensorClient",
    "LightLightnessClient",
    "LightLightnessServer",
    "LightCTLClient",
    "GatewayConfigServer",
    "GatewayConfigClient",
]

AppKeyStatus = NamedTuple(
    "ApplicationKeyStatus", [("net_key_index", int), ("app_key_index", int),]
)

NetKeyStatus = NamedTuple("NetworkKeyStatus", [("net_key_index", int),])

ModelBindStatus = NamedTuple(
    "ModelBindStatus",
    [("element_address", int), ("app_key_index", int), ("model", Type[Model]),],
)

ModelSubscriptionStatus = NamedTuple(
    "ModelSubscriptionStatus",
    [("element_address", int), ("subscription_address", int), ("model", Type[Model]),],
)

ModelSubscriptionList = NamedTuple(
    "ModelSubscriptionList",
    [("element_address", int), ("model", Type[Model]), ("addresses", Sequence[int]),],
)

ModelPublicationStatus = NamedTuple(
    "ModelSubscriptionStatus",
    [("element_address", int), ("publication_address", int), ("model", Type[Model]),],
)


class ConfigServer(Model):
    MODEL_ID = (None, 0x0000)
    OPCODES = {}  # implemented internally by BlueZ


class ConfigClient(Model):
    MODEL_ID = (None, 0x0001)
    OPCODES = {
        ConfigOpcode.APPKEY_LIST,
        ConfigOpcode.APPKEY_STATUS,
        ConfigOpcode.BEACON_STATUS,
        ConfigOpcode.COMPOSITION_DATA_STATUS,
        ConfigOpcode.DEFAULT_TTL_STATUS,
        ConfigOpcode.FRIEND_STATUS,
        ConfigOpcode.GATT_PROXY_STATUS,
        ConfigOpcode.HEARBEAT_PUBLICATION_STATUS,
        ConfigOpcode.HEARBEAT_SUBSCRIPTION_STATUS,
        ConfigOpcode.KEY_REFRESH_PHASE_STATUS,
        ConfigOpcode.LOW_POWER_NODE_POLLTIMEOUT_STATUS,
        ConfigOpcode.MODEL_APP_STATUS,
        ConfigOpcode.MODEL_PUBLICATION_STATUS,
        ConfigOpcode.MODEL_SUBSCRIPTION_STATUS,
        ConfigOpcode.NETKEY_LIST,
        ConfigOpcode.NETKEY_STATUS,
        ConfigOpcode.NETWORK_TRANSMIT_STATUS,
        ConfigOpcode.NODE_IDENTITY_STATUS,
        ConfigOpcode.NODE_RESET_STATUS,
        ConfigOpcode.RELAY_STATUS,
        ConfigOpcode.SIG_MODEL_APP_LIST,
        ConfigOpcode.SIG_MODEL_SUBSCRIPTION_LIST,
        ConfigOpcode.VENDOR_MODEL_APP_LIST,
        ConfigOpcode.VENDOR_MODEL_SUBSCRIPTION_LIST,
    }

    @staticmethod
    def _get_model_id(model):
        vendor_id, model_id = model.MODEL_ID

        if vendor_id is not None:
            return dict(vendor_id=vendor_id, model_id=model_id)

        return dict(model_id=model_id)

    async def get_param(
        self,
        nodes: Iterable[int],
        net_index: int,
        request: dict,
        status: dict,
        *,
        send_interval: float = 1.0,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_dev,
                node,
                net_index=net_index,
                opcode=request["opcode"],
                params=request["params"] or dict(),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_dev(
                node,
                net_index=0,
                opcode=status["opcode"],
                params=status["params"] or dict(),
            )
            for node in nodes
        }

        async def _progress_callback(address, result, done, total):
            if isinstance(result, TimeoutError):
                self.logger.warning("Callback timeout for addr %s", address)
                return

            try:
                aw = progress_callback(address, result["params"]["data"], done, total)
                if inspect.isawaitable(aw):
                    await aw
            except Exception as ex:
                self.logger.warning("Callback failed for addr %s: %s", address, ex)

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout,
            progress_callback=_progress_callback if progress_callback else None,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]
            for node, result in results.items()
        }

    async def get_composition_data(
        self,
        nodes: Sequence[int],
        net_index: int,
        send_interval: float = 2.0,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: float = None,
    ) -> CompositionData:
        return await self.get_param(
            nodes,
            net_index,
            request=dict(
                opcode=ConfigOpcode.COMPOSITION_DATA_GET, params=dict(page=0,)
            ),
            status=dict(
                opcode=ConfigOpcode.COMPOSITION_DATA_STATUS, params=dict(page=0,)
            ),
            send_interval=send_interval,
            progress_callback=progress_callback,
            timeout=timeout or 2 * send_interval * len(nodes),
        )

    async def get_default_ttl(
        self,
        nodes: Sequence[int],
        net_index: int,
        send_interval: float = 0.1,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: float = None,
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            net_index,
            request=dict(opcode=ConfigOpcode.DEFAULT_TTL_GET, params=dict()),
            status=dict(opcode=ConfigOpcode.DEFAULT_TTL_STATUS, params=dict()),
            send_interval=send_interval,
            progress_callback=progress_callback,
            timeout=timeout or 2 * send_interval * len(nodes),
        )

    async def get_relay(
        self,
        nodes: Sequence[int],
        net_index: int,
        send_interval: float = 0.1,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: float = None,
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            net_index,
            request=dict(opcode=ConfigOpcode.RELAY_GET, params=dict()),
            status=dict(opcode=ConfigOpcode.RELAY_STATUS, params=dict()),
            send_interval=send_interval,
            progress_callback=progress_callback,
            timeout=timeout or 2 * send_interval * len(nodes),
        )

    async def get_key_refresh_phase(
        self,
        nodes: Sequence[int],
        net_index: int,
        net_key_index: int,
        send_interval: float = 0.1,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: float = None,
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            net_index,
            request=dict(
                opcode=ConfigOpcode.KEY_REFRESH_PHASE_GET, params=dict(net_key_index=0,)
            ),
            status=dict(
                opcode=ConfigOpcode.KEY_REFRESH_PHASE_STATUS,
                params=dict(net_key_index=net_key_index,),
            ),
            send_interval=send_interval,
            progress_callback=progress_callback,
            timeout=timeout or 2 * send_interval * len(nodes),
        )

    async def add_net_key(
        self, destination: int, net_index: int, net_key_index: int, net_key: NetworkKey
    ) -> NetKeyStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETKEY_STATUS,
            params=dict(net_key_index=net_key_index, status=StatusCode.SUCCESS,),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETKEY_ADD,
            params=dict(net_key_index=net_key_index, net_key=net_key.bytes,),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot add net key", status)

    async def add_app_key(
        self,
        destination: int,
        net_index: int,
        app_key_index: int,
        net_key_index: int,
        app_key: ApplicationKey,
    ) -> AppKeyStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.APPKEY_STATUS,
            params=dict(net_key_index=net_key_index, app_key_index=app_key_index,),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.APPKEY_ADD,
            params=dict(
                net_key_index=net_key_index,
                app_key_index=app_key_index,
                app_key=app_key.bytes,
            ),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot add app key", status)

        return AppKeyStatus(
            status["params"]["net_key_index"], status["params"]["app_key_index"]
        )

    async def update_app_key(
        self,
        destination: int,
        net_index: int,
        net_key_index: int,
        app_key_index: int,
        app_key: ApplicationKey,
    ) -> AppKeyStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.APPKEY_STATUS,
            params=dict(net_key_index=net_key_index, app_key_index=app_key_index,),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.APPKEY_UPDATE,
            params=dict(app_key_index=0, net_key_index=0, app_key=app_key.bytes,),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot update app key", status)

        return AppKeyStatus(
            status["params"]["net_key_index"], status["params"]["app_key_index"]
        )

    async def bind_app_key(
        self,
        destination: int,
        net_index: int,
        element_address: int,
        app_key_index: int,
        model: Type[Model],
    ) -> ModelBindStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_APP_STATUS,
            params=dict(
                element_address=element_address,
                app_key_index=app_key_index,
                model=self._get_model_id(model),
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_APP_BIND,
            params=dict(
                element_address=element_address,
                app_key_index=app_key_index,
                model=self._get_model_id(model),
            ),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot bind app key", status)

        return ModelBindStatus(
            status["params"]["element_address"],
            status["params"]["app_key_index"],
            model,
        )

    async def get_network_transmission(
        self, destination: int, net_index: int
    ) -> Tuple[int, int]:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETWORK_TRANSMIT_STATUS,
            params=dict(),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETWORK_TRANSMIT_GET,
            params=dict(),
        )

        status = await self.query(request, status)
        return (status["params"]["interval"], status["params"]["count"])

    async def set_network_transmission(
        self, destination: int, net_index: int, interval: int, count: int
    ) -> Tuple[int, int]:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETWORK_TRANSMIT_STATUS,
            params=dict(interval=interval, count=count),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETWORK_TRANSMIT_SET,
            params=dict(interval=interval, count=count),
        )

        status = await self.query(request, status)
        return (status["params"]["interval"], status["params"]["count"])

    async def update_net_key(
        self, destination: int, net_index: int, net_key_index: int, net_key: NetworkKey
    ) -> int:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETKEY_STATUS,
            params=dict(net_key_index=net_key_index,),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.NETKEY_UPDATE,
            params=dict(net_key_index=0, net_key=net_key.bytes,),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot update net key", status)

        return status["params"]["net_key_index"]

    async def add_subscription(
        self,
        destination: int,
        net_index: int,
        element_address: int,
        subscription_address: int,
        model: Type[Model],
    ) -> ModelSubscriptionStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_SUBSCRIPTION_STATUS,
            params=dict(
                element_address=element_address,
                address=subscription_address,
                model=self._get_model_id(model),
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_SUBSCRIPTION_ADD,
            params=dict(
                element_address=element_address,
                address=subscription_address,
                model=self._get_model_id(model),
            ),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot add subscription", status)

        return ModelSubscriptionStatus(
            status["params"]["element_address"], status["params"]["address"], model
        )

    async def del_subscription(
        self,
        destination: int,
        net_index: int,
        element_address: int,
        subscription_address: int,
        model: Type[Model],
    ) -> ModelSubscriptionStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_SUBSCRIPTION_STATUS,
            params=dict(
                element_address=element_address,
                address=subscription_address,
                model=self._get_model_id(model),
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_SUBSCRIPTION_DELETE,
            params=dict(
                element_address=element_address,
                address=subscription_address,
                model=self._get_model_id(model),
            ),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot add subscription", status)

        return ModelSubscriptionStatus(
            status["params"]["element_address"], status["params"]["address"], model
        )

    async def clear_subscriptions(
        self, destination: int, net_index: int, element_address: int, model: Type[Model]
    ) -> ModelSubscriptionStatus:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_SUBSCRIPTION_STATUS,
            params=dict(
                element_address=element_address,
                address=0,
                model=self._get_model_id(model),
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_SUBSCRIPTION_DELETE_ALL,
            params=dict(
                element_address=element_address, model=self._get_model_id(model),
            ),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot add subscription", status)

        return ModelSubscriptionStatus(
            status["params"]["element_address"], status["params"]["address"], model
        )

    async def get_subscriptions(
        self, destination: int, net_index: int, element_address: int, model: Type[Model]
    ) -> ModelSubscriptionList:
        if model.MODEL_ID[0] is not None:
            status_opcode = ConfigOpcode.VENDOR_MODEL_SUBSCRIPTION_LIST
            request_opcode = ConfigOpcode.VENDOR_MODEL_SUBSCRIPTION_GET
        else:
            status_opcode = ConfigOpcode.SIG_MODEL_SUBSCRIPTION_LIST
            request_opcode = ConfigOpcode.SIG_MODEL_SUBSCRIPTION_GET

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=status_opcode,
            params=dict(
                element_address=element_address, model=self._get_model_id(model),
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=request_opcode,
            params=dict(
                element_address=element_address, model=self._get_model_id(model),
            ),
        )

        status = await self.query(request, status)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot get subscription list", status)

        return ModelSubscriptionList(
            status["params"]["element_address"], model, status["params"]["addresses"],
        )

    async def add_publication(
        self,
        destination: int,
        net_index: int,
        element_address: int,
        publication_address: int,
        app_key_index: int,
        model: Type[Model],
        TTL: int = 8,
        publish_step_resolution: int = 2,
        publish_number_of_steps: int = 6,  # 60seconds
        retransmit_count: int = 0,
        retransmit_interval: int = 50,
    ) -> ModelSubscriptionStatus:

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_PUBLICATION_STATUS,
            params=dict(
                status=StatusCode.SUCCESS,
                element_address=element_address,
                publish_address=publication_address,
                TTL=TTL,
                app_key_index=app_key_index,
                credential_flag=PublishFriendshipCredentialsFlag.MASTER_SECURITY,
                RFU=0,
                publish_period=dict(
                    step_resolution=publish_step_resolution,
                    number_of_steps=publish_number_of_steps,
                ),
                retransmit=dict(count=retransmit_count, interval=retransmit_interval),
                model=self._get_model_id(model),
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.MODEL_PUBLICATION_SET,
            params=dict(
                element_address=element_address,
                publish_address=publication_address,
                TTL=TTL,
                app_key_index=app_key_index,
                credential_flag=PublishFriendshipCredentialsFlag.MASTER_SECURITY,
                RFU=0,
                publish_period=dict(
                    step_resolution=publish_step_resolution,
                    number_of_steps=publish_number_of_steps,
                ),
                retransmit=dict(count=retransmit_count, interval=retransmit_interval),
                model=self._get_model_id(model),
            ),
        )

        status = await self.query(request, status, send_interval=0.2, timeout=4)

        if status["params"]["status"] != StatusCode.SUCCESS:
            raise ModelOperationError("Cannot add subscription", status)

        return ModelPublicationStatus(
            status["params"]["element_address"],
            status["params"]["publish_address"],
            model,
        )

    async def set_beacon(
        self,
        destination: int,
        net_index: int,
        enabled: bool,
        send_interval: float = 2.0,
        timeout: float = 5.0,
    ) -> CompositionData:
        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.BEACON_STATUS,
            params=dict(
                beacon=SecureNetworkBeacon.ON if enabled else SecureNetworkBeacon.OFF,
            ),
        )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=ConfigOpcode.BEACON_SET,
            params=dict(
                beacon=SecureNetworkBeacon.ON if enabled else SecureNetworkBeacon.OFF,
            ),
        )

        status = await self.query(
            request, status, send_interval=send_interval, timeout=timeout
        )

        return status["params"]["beacon"]


class HealthServer(Model):
    MODEL_ID = (None, 0x0002)
    OPCODES = {
        HealthOpcode.FAULT_GET,
        HealthOpcode.FAULT_CLEAR,
        HealthOpcode.FAULT_CLEAR_UNACKNOWLEDGED,
        HealthOpcode.FAULT_TEST,
        HealthOpcode.FAULT_TEST_UNACKNOWLEDGED,
        HealthOpcode.PERIOD_GET,
        HealthOpcode.PERIOD_SET,
        HealthOpcode.PERIOD_SET_UNACKNOWLEDGED,
        HealthOpcode.ATTENTION_GET,
        HealthOpcode.ATTENTION_SET,
        HealthOpcode.ATTENTION_SET_UNACKNOWLEDGED,
    }


class HealthClient(Model):
    MODEL_ID = (None, 0x0003)
    OPCODES = {
        HealthOpcode.CURRENT_STATUS,
        HealthOpcode.FAULT_STATUS,
        HealthOpcode.PERIOD_STATUS,
        HealthOpcode.ATTENTION_STATUS,
    }

    async def attention(self, destination: int, app_index: int, attention: int) -> int:
        status = self.expect_app(
            destination,
            app_index=app_index,
            destination=None,
            opcode=HealthOpcode.ATTENTION_STATUS,
            params=dict(attention=attention,),
        )

        request = partial(
            self.send_app,
            destination,
            app_index=app_index,
            opcode=HealthOpcode.ATTENTION_SET,
            params=dict(attention=attention,),
        )

        status = await self.query(request, status)
        return status["params"]["attention"]

    async def attention_unack(self, destination: int, app_index: int, attention: int):
        request = partial(
            self.send_app,
            destination,
            app_index=app_index,
            opcode=HealthOpcode.ATTENTION_SET_UNACKNOWLEDGED,
            params=dict(attention=attention,),
        )

        await self.repeat(request)


class DebugServer(Model):
    MODEL_ID = (0x0136, 0x0002)
    OPCODES = {
        DebugOpcode.OPCODE,
    }


class DebugClient(Model):
    MODEL_ID = (0x0136, 0x0016)
    OPCODES = {
        DebugOpcode.OPCODE,
    }

    async def get_param(
        self,
        nodes: Iterable[int],
        net_index: int,
        request: int,
        status: int,
        *,
        send_interval: float = 1.0,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_dev,
                node,
                net_index=net_index,
                opcode=DebugOpcode.OPCODE,
                params=dict(subopcode=request),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_dev(
                node,
                net_index=0,
                opcode=DebugOpcode.OPCODE,
                params=dict(subopcode=status),
            )
            for node in nodes
        }

        async def _progress_callback(address, result, done, total):
            if isinstance(result, TimeoutError):
                self.logger.warning("Callback timeout for addr %s", address)
                return

            try:
                aw = progress_callback(address, result["params"]["data"], done, total)

                if inspect.isawaitable(aw):
                    await aw
            except Exception as ex:
                self.logger.warning("Callback failed for addr %s: %s", address, ex)

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout,
            progress_callback=_progress_callback if progress_callback else None,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]["data"]
            for node, result in results.items()
        }

    async def get_uptime(
        self, nodes: Sequence[int], app_index: int
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            app_index,
            DebugSubOpcode.UPTIME_GET,
            DebugSubOpcode.UPTIME_STATUS,
            send_interval=0.1,
            timeout=len(nodes) * 0.5,
        )

    async def get_last_sw_fault(
        self, nodes: Sequence[int], app_index: int
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            app_index,
            DebugSubOpcode.LAST_SW_FAULT_GET,
            DebugSubOpcode.LAST_SW_FAULT_STATUS,
            send_interval=0.1,
            timeout=len(nodes) * 0.5,
        )

    async def get_firmware_version(
        self, nodes: Sequence[int], app_index: int
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            app_index,
            DebugSubOpcode.FULL_FIRMWARE_VERSION_GET,
            DebugSubOpcode.FULL_FIRMWARE_VERSION_STATUS,
            send_interval=0.1,
            timeout=len(nodes) * 0.5,
        )

    async def get_app_version(
        self, nodes: Sequence[int], app_index: int
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            app_index,
            DebugSubOpcode.PROVISIONED_APP_VERSION_GET,
            DebugSubOpcode.PROVISIONED_APP_VERSION_STATUS,
            send_interval=0.1,
            timeout=len(nodes) * 0.5,
        )

    async def get_ivindex(
        self, nodes: Sequence[int], app_index: int
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            app_index,
            DebugSubOpcode.IV_INDEX_GET,
            DebugSubOpcode.IV_INDEX_STATUS,
            send_interval=0.1,
            timeout=len(nodes) * 0.5,
        )

    async def get_system_stats(
        self, nodes: Sequence[int], app_index: int
    ) -> Dict[int, Optional[Any]]:
        return await self.get_param(
            nodes,
            app_index,
            DebugSubOpcode.SYSTEM_STATS_GET,
            DebugSubOpcode.SYSTEM_STATS_STATUS,
            send_interval=2,
            timeout=max(5, len(nodes) * 2),
        )

    async def get_arap_content(
        self,
        nodes: Sequence[int],
        app_index: int,
        *,
        progress_callback: Optional[ProgressCallback] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_dev,
                node,
                net_index=app_index,
                opcode=DebugOpcode.OPCODE,
                params=dict(
                    subopcode=DebugSubOpcode.ARAP_LIST_CONTENT_GET, data=dict(page=0),
                ),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_dev(
                node,
                net_index=0,
                opcode=DebugOpcode.OPCODE,
                params=dict(subopcode=DebugSubOpcode.ARAP_LIST_CONTENT_STATUS),
            )
            for node in nodes
        }

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=0.5,
            timeout=max(2.5, len(nodes) * 1.0),
            progress_callback=progress_callback,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]["data"]
            for node, result in results.items()
        }


class NetworkDiagnosticClient(Model):
    MODEL_ID = (0x0136, 0x0014)
    OPCODES = {
        NetworkDiagnosticServerOpcode.OPCODE,
    }


class NetworkDiagnosticSetupClient(Model):
    MODEL_ID = (0x0136, 0x0015)
    OPCODES = {
        NetworkDiagnosticSetupServerOpcode.OPCODE,
    }


class GenericOnOffServer(Model):
    MODEL_ID = (None, 0x1000)
    OPCODES = {
        GenericOnOffOpcode.ONOFF_GET,
        GenericOnOffOpcode.ONOFF_SET,
        GenericOnOffOpcode.ONOFF_SET_UNACKNOWLEDGED,
    }


class GenericOnOffClient(Model):
    MODEL_ID = (None, 0x1001)
    OPCODES = {
        GenericOnOffOpcode.ONOFF_STATUS,
    }
    _tid = 0

    @property
    def tid(self) -> int:
        self._tid = (self._tid + 1) % 255
        return self._tid

    async def set_onoff(
        self,
        destination: int,
        app_index: int,
        onoff: int,
        delay: float = 0.5,
        send_interval: float = 0.07,
    ) -> int:
        values = dict(delay=delay, send_interval=send_interval, tid=self.tid)

        status = self.expect_app(
            destination,
            app_index=app_index,
            destination=None,
            opcode=GenericOnOffOpcode.ONOFF_STATUS,
            params=dict(present_onoff=onoff),
        )

        async def request():
            ret = self.send_app(
                destination,
                app_index=app_index,
                opcode=GenericOnOffOpcode.ONOFF_SET,
                params=dict(
                    onoff=onoff,
                    tid=values["tid"],
                    transition_time=0,
                    delay=values["delay"],
                ),
            )
            values["delay"] -= values["send_interval"]
            values["delay"] = max(0, values["delay"])

            return await ret

        status = await self.query(
            request, status, send_interval=values["send_interval"], timeout=1
        )

        return status["params"]["present_onoff"]

    async def set_onoff_unack(
        self,
        destination: int,
        app_index: int,
        onoff: int,
        delay: float = 0.5,
        send_interval: float = 0.07,
        transition_time: float = 0,
        retransmissions: int = 6,
    ):
        values = dict(delay=delay, send_interval=send_interval, tid=self.tid)

        async def request():
            ret = self.send_app(
                destination,
                app_index=app_index,
                opcode=GenericOnOffOpcode.ONOFF_SET_UNACKNOWLEDGED,
                params=dict(
                    onoff=onoff,
                    tid=values["tid"],
                    transition_time=transition_time,
                    delay=values["delay"],
                ),
            )
            values["delay"] -= values["send_interval"]
            values["delay"] = max(0, values["delay"])

            return await ret

        await self.repeat(
            request,
            retransmissions=retransmissions,
            send_interval=values["send_interval"],
        )

    async def get_light_status(
        self,
        nodes: Sequence[int],
        app_index: int,
        *,
        send_interval: float = 0.1,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_app,
                node,
                app_index=app_index,
                opcode=GenericOnOffOpcode.ONOFF_GET,
                params=dict(),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_app(
                node,
                app_index=0,
                destination=None,
                opcode=GenericOnOffOpcode.ONOFF_STATUS,
                params=dict(),
            )
            for node in nodes
        }

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout or len(nodes) * 0.5,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]
            for node, result in results.items()
        }


class SceneClient(Model):
    MODEL_ID = (None, 0x1205)
    OPCODES = {
        SceneOpcode.SCENE_GET,
        SceneOpcode.SCENE_RECALL,
        SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED,
        SceneOpcode.SCENE_STATUS,
        SceneOpcode.SCENE_REGISTER_GET,
        SceneOpcode.SCENE_REGISTER_STATUS,
        SceneOpcode.SCENE_STORE,
        SceneOpcode.SCENE_STORE_UNACKNOWLEDGED,
        SceneOpcode.SCENE_DELETE,
        SceneOpcode.SCENE_DELETE_UNACKNOWLEDGED,
    }

    def __init__(self, element: "Element"):
        super().__init__(element)
        self.__tid = itertools.cycle(range(255))

    async def recall_scene_unack(
        self,
        destination: int,
        app_index: int,
        scene_number: int,
        transition_time: float,
    ):
        values = dict(delay=0.5, send_interval=0.075, tid=next(self.__tid))

        async def request():
            ret = self.send_app(
                destination,
                app_index=app_index,
                opcode=SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED,
                params=dict(
                    scene_number=scene_number,
                    tid=values["tid"],
                    transition_time=transition_time,
                    delay=values["delay"],
                ),
            )
            values["delay"] -= values["send_interval"]
            values["delay"] = max(0, values["delay"])

            return await ret

        await self.repeat(
            request, retransmissions=6, send_interval=values["send_interval"]
        )


class GenericLevelClient(Model):
    MODEL_ID = (None, 0x1003)
    OPCODES = {
        GenericLevelOpcode.LEVEL_GET,
        GenericLevelOpcode.LEVEL_SET,
        GenericLevelOpcode.LEVEL_SET_UNACKNOWLEDGED,
        GenericLevelOpcode.DELTA_SET,
        GenericLevelOpcode.DELTA_SET_UNACKNOWLEDGED,
        GenericLevelOpcode.MOVE_SET,
        GenericLevelOpcode.MOVE_SET_UNACKNOWLEDGED,
    }

    def __init__(self, element: "Element"):
        super().__init__(element)
        self.__tid = itertools.cycle(range(255))

    async def set_level_unack(
        self,
        destination: int,
        app_index: int,
        level: int,
        delay: float = 0.5,
        send_interval: float = 0.07,
        transition_time: float = 0,
        retransmissions: int = 6,
    ):

        values = dict(delay=delay, send_interval=send_interval, tid=next(self.__tid))

        async def request():
            ret = self.send_app(
                destination,
                app_index=app_index,
                opcode=GenericLevelOpcode.LEVEL_SET_UNACKNOWLEDGED,
                params=dict(
                    level=level,
                    tid=values["tid"],
                    transition_time=transition_time,
                    delay=values["delay"],
                ),
            )
            values["delay"] -= values["send_interval"]
            values["delay"] = max(0, values["delay"])

            return await ret

        await self.repeat(
            request,
            retransmissions=retransmissions,
            send_interval=values["send_interval"],
        )


class LightLightnessServer(Model):
    MODEL_ID = (None, 0x1300)
    OPCODES = {
        LightLightnessOpcode.LIGHTNESS_GET,
        LightLightnessOpcode.LIGHTNESS_SET,
        LightLightnessOpcode.LIGHTNESS_SET_UNACKNOWLEDGED,
        LightLightnessOpcode.LIGHTNESS_STATUS,
    }


class LightLightnessSetupServer(Model):
    MODEL_ID = (None, 0x1301)
    OPCODES = {
        LightLightnessSetupOpcode.LIGHTNESS_DEFAULT_SET,
        LightLightnessSetupOpcode.LIGHTNESS_DEFAULT_SET_UNACKNOWLEDGED,
        LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET,
        LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET_UNACKNOWLEDGED,
    }


class LightLightnessClient(Model):
    MODEL_ID = (None, 0x1302)
    OPCODES = {
        LightLightnessOpcode.LIGHTNESS_GET,
        LightLightnessOpcode.LIGHTNESS_SET,
        LightLightnessOpcode.LIGHTNESS_SET_UNACKNOWLEDGED,
        LightLightnessOpcode.LIGHTNESS_STATUS,
        LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET_UNACKNOWLEDGED,
    }
    _tid = 0

    @property
    def tid(self) -> int:
        self._tid = (self._tid + 1) % 255
        return self._tid

    async def set_lightness_range_unack(
        self,
        destination: int,
        app_index: int,
        min_lightness: int,
        max_lightness: int,
        *,
        retransmissions: int = 6,
        send_interval: float = 0.075,
    ) -> None:
        async def request():
            ret = self.send_app(
                destination,
                app_index=app_index,
                opcode=LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET_UNACKNOWLEDGED,
                params=dict(
                    range_min=min_lightness,
                    range_max=max_lightness,
                ),
            )

            return await ret

        await self.repeat(request, retransmissions=retransmissions, send_interval=send_interval)

    async def get_lightness(
        self,
        nodes: Sequence[int],
        app_index: int,
        *,
        send_interval: float = 0.1,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_app,
                node,
                app_index=app_index,
                opcode=LightLightnessOpcode.LIGHTNESS_GET,
                params=dict(),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_app(
                node,
                app_index=0,
                destination=None,
                opcode=LightLightnessOpcode.LIGHTNESS_STATUS,
                params=dict(),
            )
            for node in nodes
        }

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout or len(nodes) * 0.5,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]
            for node, result in results.items()
        }

    async def set_lightness(
        self,
        nodes: Sequence[int],
        lightness: int,
        app_index: int,
        *,
        delay: float = 0.5,
        send_interval: float = 0.1,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        values = dict(delay=delay, send_interval=send_interval, tid=self.tid)

        requests = {
            node: partial(
                self.send_app,
                node,
                app_index=app_index,
                opcode=LightLightnessOpcode.LIGHTNESS_SET,
                params=dict(lightness=lightness, tid=values["tid"],),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_app(
                node,
                app_index=0,
                destination=None,
                opcode=LightLightnessOpcode.LIGHTNESS_STATUS,
                params=dict(),
            )
            for node in nodes
        }

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout or len(nodes) * 0.5,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]
            for node, result in results.items()
        }


class SensorClient(Model):
    MODEL_ID = (None, 0x1102)
    OPCODES = {
        SensorOpcode.SENSOR_DESCRIPTOR_GET,
        SensorOpcode.SENSOR_DESCRIPTOR_STATUS,
        SensorOpcode.SENSOR_GET,
        SensorOpcode.SENSOR_STATUS,
    }


class LightCTLClient(Model):
    MODEL_ID = (None, 0x1305)
    OPCODES = {
        LightCTLOpcode.CTL_GET,
        LightCTLOpcode.CTL_SET,
        LightCTLOpcode.CTL_SET_UNACKNOWLEDGED,
        LightCTLOpcode.CTL_STATUS,
        LightCTLOpcode.CTL_TEMPERATURE_GET,
        LightCTLOpcode.CTL_TEMPERATURE_RANGE_GET,
        LightCTLOpcode.CTL_TEMPERATURE_RANGE_STATUS,
        LightCTLOpcode.CTL_TEMPERATURE_SET,
        LightCTLOpcode.CTL_TEMPERATURE_SET_UNACKNOWLEDGED,
        LightCTLOpcode.CTL_TEMPERATURE_STATUS,
        LightCTLOpcode.CTL_TEMPERATURE_DEFAULT_GET,
        LightCTLOpcode.CTL_TEMPERATURE_DEFAULT_STATUS,
    }
    _tid = 0

    @property
    def tid(self) -> int:
        self._tid = (self._tid + 1) % 255
        return self._tid

    async def get_ctl(
        self,
        nodes: Sequence[int],
        app_index: int,
        *,
        send_interval: float = 0.1,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_app,
                node,
                app_index=app_index,
                opcode=LightCTLOpcode.CTL_TEMPERATURE_GET,
                params=dict(),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_app(
                node,
                app_index=0,
                destination=None,
                opcode=LightCTLOpcode.CTL_TEMPERATURE_STATUS,
                params=dict(),
            )
            for node in nodes
        }

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout or len(nodes) * 0.5,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]
            for node, result in results.items()
        }

    async def set_ctl(
        self,
        nodes: Sequence[int],
        app_index: int,
        *,
        ctl_temperature: int,
        send_interval: float = 0.1,
        timeout: Optional[float] = None,
    ) -> Dict[int, Optional[Any]]:
        requests = {
            node: partial(
                self.send_app,
                node,
                app_index=app_index,
                opcode=LightCTLOpcode.CTL_TEMPERATURE_SET,
                params=dict(
                    ctl_temperature=ctl_temperature, ctl_delta_uv=0, tid=self.tid
                ),
            )
            for node in nodes
        }

        statuses = {
            node: self.expect_app(
                node,
                app_index=0,
                destination=None,
                opcode=LightCTLOpcode.CTL_TEMPERATURE_STATUS,
                params=dict(),
            )
            for node in nodes
        }

        results = await self.bulk_query(
            requests,
            statuses,
            send_interval=send_interval,
            timeout=timeout or len(nodes) * 0.5,
        )

        return {
            node: None if isinstance(result, Exception) else result["params"]
            for node, result in results.items()
        }


class GatewayConfigServer(Model):
    MODEL_ID = (0x0136, 0x001D)
    OPCODES = {
        GatewayConfigServerOpcode.OPCODE,
    }

    def send_configuration_status(
        self,
        destination: int,
        net_index: int,
        chip_rev_id: int,
        mtu: int,
        mac: str,
        server: Tuple[str, int],
        reconnect: int,
        ip: str,
        dns: str,
        gateway: str,
        netmask: int,
        dhcp: DhcpFlag,
    ):
        params = dict(
            subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
            payload=dict(
                chip_revision_id=chip_rev_id,
                mtu_size=mtu,
                mac_addr=mac,
                server_port_number=server[1],
                reconnect_interval=reconnect,
                server_address_length=len(server[0]),
                server_address=server[0],
                dns_ip_address=dns,
                ip_addr=ip,
                gateway_ip_addr=gateway,
                netmask=netmask,
                flags=dhcp,
            ),
        )
        self.send_dev(destination, net_index, GatewayConfigServerOpcode.OPCODE, params)

    def send_packets_status(
        self,
        destination: int,
        net_index: int,
        rx_errors: int,
        tx_errors: int,
        bandwidth: int,
        state: BitStruct(),
    ):
        params = dict(
            subopcode=GatewayConfigServerSubOpcode.GATEWAY_PACKETS_STATUS,
            payload=dict(
                total_eth_rx_errors=rx_errors,
                total_eth_tx_errors=tx_errors,
                bandwidth=bandwidth,
                connection_state=state,
            ),
        )
        self.send_dev(destination, net_index, GatewayConfigServerOpcode.OPCODE, params)


class GatewayConfigClient(Model):
    MODEL_ID = (0x0136, 0x001E)
    OPCODES = {
        GatewayConfigServerOpcode.OPCODE,
    }

    async def configuration_get(self, destination: int, net_index: int):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_GET,
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def configuration_set(
        self,
        destination: int,
        net_index: int,
        mtu: int,
        mac: str,
        server: Tuple[str, int],
        reconnect: int,
        dns: str = None,
        ip: str = None,
        gateway: str = None,
        netmask: int = None,
    ):
        payload = dict(
            mtu_size=mtu,
            mac_address=mac,
            server_port_number=server[1],
            reconnect_interval=reconnect,
            server_address_length=len(server[0]),
            server_address=server[0],
        )

        if dns:
            payload.update(dict(dns_ip_address=dns,))

        if ip and gateway and netmask:
            payload.update(
                dict(ip_address=ip, gateway_ip_address=gateway, netmask=netmask,)
            )

        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
                payload=payload,
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def packets_get(self, destination: int, net_index: int):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(subopcode=GatewayConfigServerSubOpcode.GATEWAY_PACKETS_GET,),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(subopcode=GatewayConfigServerSubOpcode.GATEWAY_PACKETS_STATUS,),
        )
        return await self.query(request, status, timeout=1.0)

    async def packets_clear(self, destination: int, net_index: int):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(subopcode=GatewayConfigServerSubOpcode.GATEWAY_PACKETS_CLEAR,),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_PACKETS_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def mtu_set(self, destination: int, net_index: int, mtu: int):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.MTU_SIZE_SET,
                payload=dict(mtu_size=mtu),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def mac_set(self, destination: int, net_index: int, mac: str):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.ETHERNET_MAC_ADDRESS_SET,
                payload=dict(mac_address=mac),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def server_set(
        self, destination: int, net_index: int, server: Tuple[str, int],
    ):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.SERVER_ADDRESS_AND_PORT_NUMBER_SET,
                payload=dict(
                    server_port_number=server[1],
                    server_address_length=len(server[0]),
                    server_address=server[0],
                ),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def reconnect_set(self, destination: int, net_index: int, reconnect: int):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.RECONNECT_INTERVAL_SET,
                payload=dict(reconnect_interval=reconnect),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def dns_set(self, destination: int, net_index: int, dns: str):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.DNS_IP_ADDRESS_SET,
                payload=dict(dns_ip_address=dns),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def ip_set(self, destination: int, net_index: int, ip: str):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.IP_ADDRESS_SET,
                payload=dict(ip_address=ip),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def gateway_set(self, destination: int, net_index: int, gateway: str):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_IP_ADDRESS_SET,
                payload=dict(gateway_ip_address=gateway),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)

    async def netmask_set(self, destination: int, net_index: int, netmask: int):
        request = partial(
            self.send_dev,
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.NETMASK_SET,
                payload=dict(netmask=netmask),
            ),
        )

        status = self.expect_dev(
            destination,
            net_index=net_index,
            opcode=GatewayConfigServerOpcode.OPCODE,
            params=dict(
                subopcode=GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
                payload=...,
            ),
        )
        return await self.query(request, status, timeout=1.0)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import enum
import math
import operator
from uuid import UUID

import bitstring
from crc.crc import Configuration, CrcCalculator

from bluetooth_mesh.crypto import (
    ApplicationKey,
    NetworkKey,
    aes_ccm_decrypt,
    aes_ccm_encrypt,
    aes_cmac,
    aes_ecb,
)
from bluetooth_mesh.provisioning import (
    BearerOpcode,
    ProvisioningBearerControl,
    ProvisioningPDU,
    ProvisioningPDUType,
    TransactionContinuationPDU,
    TransactionPDUSegment,
    TransactionStartPDU,
)


class BeaconType(enum.Enum):
    UNPROVISIONED_DEVICE = 0x00
    SECURE_NETWORK = 0x01


class UnprovisionedDeviceBeacon:
    BEACON_FORMAT = "bytes:16, uint:16, bytes"

    def __init__(self, uuid, oob, uri_hash=None):
        self.uuid = uuid
        self.oob = oob

        if uri_hash and len(uri_hash) != 4:
            raise ValueError("Wrong size of URI hash, expected 4 bytes")

        self.uri_hash = uri_hash

    def __str__(self):
        return "<%s: uuid=%s, oob=%x, uri_hash=%s>" % (
            type(self).__name__,
            self.uuid,
            self.oob,
            self.uri_hash,
        )

    @classmethod
    def unpack(cls, beacon):
        uuid, oob, uri_hash = bitstring.BitString(beacon).unpack(cls.BEACON_FORMAT)

        if uri_hash and len(uri_hash) != 4:
            raise ValueError("Wrong size of URI hash, expected 4 bytes")

        return cls(UUID(bytes=uuid), oob, uri_hash or None)

    def pack(self):
        beacon = bitstring.pack(
            self.BEACON_FORMAT, self.uuid.bytes, self.oob, self.uri_hash or b""
        ).bytes
        return beacon


class SecureNetworkBeacon:
    BEACON_FORMAT = "pad:6, uint:1, uint:1, bytes:8, uintbe:32"
    BEACON_AUTH_SIZE = 8

    def __init__(self, key_refresh, iv_update, iv_index, network_id):
        self.key_refresh = key_refresh
        self.iv_update = iv_update
        self.iv_index = iv_index
        self.network_id = network_id

    def __str__(self):
        return "<%s: key_refresh=%s, iv_update=%s, ivindex=%d, network_id=%s>" % (
            type(self).__name__,
            self.key_refresh,
            self.iv_update,
            self.iv_index,
            self.network_id.hex(),
        )

    @classmethod
    def unpack(cls, message):
        beacon, auth = (
            message[: -cls.BEACON_AUTH_SIZE],
            message[-cls.BEACON_AUTH_SIZE :],
        )

        iv_update, key_refresh, network_id, iv_index = bitstring.BitString(
            beacon
        ).unpack(cls.BEACON_FORMAT)

        return cls(bool(key_refresh), bool(iv_update), iv_index, network_id), auth

    def pack(self, network_key):
        beacon = bitstring.pack(
            self.BEACON_FORMAT,
            self.iv_update,
            self.key_refresh,
            self.network_id,
            self.iv_index,
        ).bytes
        auth = aes_cmac(network_key.beacon_key, beacon)[: self.BEACON_AUTH_SIZE]
        return beacon, auth

    def verify(self, auth, network_key):
        _, _auth = self.pack(network_key)
        return auth == _auth


class Nonce:
    def __init__(self, src, dst, ttl, ctl):
        super().__init__()
        self.src = src
        self.dst = dst
        self.ttl = ttl
        self.ctl = ctl

    def network(self, seq, iv_index):
        return bitstring.pack(
            "uint:8, uint:1, uint:7, uintbe:24, uintbe:16, pad:16, uintbe:32",
            0x00,
            self.ctl,
            self.ttl,
            seq,
            self.src,
            iv_index,
        ).bytes

    def application(self, seq, iv_index, szmic=False):
        return bitstring.pack(
            "uint:8, uint:1, pad:7, uintbe:24, uintbe:16, uintbe:16, uintbe:32",
            0x01,
            szmic,
            seq,
            self.src,
            self.dst,
            iv_index,
        ).bytes

    def device(self, seq, iv_index, szmic=False):
        return bitstring.pack(
            "uint:8, uint:1, pad:7, uintbe:24, uintbe:16, uintbe:16, uintbe:32",
            0x02,
            szmic,
            seq,
            self.src,
            self.dst,
            iv_index,
        ).bytes

    def proxy(self, seq, iv_index):
        return bitstring.pack(
            "uint:8, pad:8, uintbe:24, uintbe:16, uintbe:16, uintbe:32",
            0x03,
            seq,
            self.src,
            self.dst,
            iv_index,
        ).bytes

    def solicitation(self, seq):
        return bitstring.pack(
            "uint:8, pad:8, uintbe:24, uintbe:16, pad:16, uintbe:32",
            0x04,
            seq,
            self.src,
            0,
        ).bytes


class Segment:
    MAX_TRANSPORT_PDU = 15
    SEGMENT_SIZE = 12

    def __init__(self, src, dst, ttl, ctl, payload):
        self.src = src
        self.dst = dst
        self.ttl = ttl
        self.ctl = ctl
        self.payload = payload
        self.nonce = Nonce(self.src, self.dst, self.ttl, self.ctl)

    def __eq__(self, other):
        if not isinstance(other, Segment):
            raise NotImplementedError
        return (
            self.src == other.src
            and self.dst == other.dst
            and self.ttl == other.ttl
            and self.ctl == other.ctl
            and self.payload == other.payload
        )

    def get_opcode(self, application_key):
        raise NotImplementedError

    def _segments(self, application_key, seq, payload, szmic):
        opcode = self.get_opcode(application_key)
        seq_zero = seq & 0x1FFF
        seg = len(payload) > self.MAX_TRANSPORT_PDU

        if seg:
            segments = list(
                payload[i : i + self.SEGMENT_SIZE]
                for i in range(0, len(payload), self.SEGMENT_SIZE)
            )

            seg_n = len(segments) - 1

            for seg_o, segment in enumerate(segments):
                yield bitstring.pack(
                    "uint:1, bits:7, uint:1, uint:13, uint:5, uint:5, bytes",
                    seg,
                    opcode,
                    szmic,
                    seq_zero,
                    seg_o,
                    seg_n,
                    segment,
                ).bytes
        else:
            yield bitstring.pack("uint:1, bits:7, bytes", seg, opcode, payload).bytes


class AccessMessage(Segment):
    def __init__(self, src, dst, ttl, payload):
        super().__init__(src, dst, ttl, False, payload)

    def get_opcode(self, application_key):
        akf = isinstance(application_key, ApplicationKey)
        aid = application_key.aid
        return bitstring.pack("uint:1, uint:6", akf, aid)

    def segments(self, application_key, seq, iv_index, szmic=False):
        short_mic_len = len(self.payload) + 4
        long_mic_len = len(self.payload) + 8

        # Use large MIC if it doesn't affect segmentation
        if len(self.payload) >= self.SEGMENT_SIZE and len(self.payload) < 376:
            szmic = szmic or (
                math.ceil(short_mic_len / self.SEGMENT_SIZE)
                == math.ceil(long_mic_len / self.SEGMENT_SIZE)
            )

        akf = isinstance(application_key, ApplicationKey)
        nonce = (self.nonce.application if akf else self.nonce.device)(
            seq, iv_index, szmic
        )

        upper_transport_pdu = aes_ccm_encrypt(
            application_key.bytes, nonce, self.payload, b"", 8 if szmic else 4
        )

        yield from super()._segments(
            application_key, seq, payload=upper_transport_pdu, szmic=szmic
        )

    @classmethod
    def decrypt(cls, app_key, iv_index, ctl, ttl, seq, src, dst, transport_pdu):
        seg, akf, aid = bitstring.BitString(transport_pdu).unpack(
            "uint:1, uint:1, uint:6"
        )

        # works only for unsegmented messages!
        if seg:
            raise NotImplementedError
        if app_key.aid != aid:
            raise KeyError
        transport_nonce = Nonce(src, dst, ttl, ctl)
        nonce = (transport_nonce.application if akf else transport_nonce.device)(
            seq, iv_index
        )
        decrypted_access = aes_ccm_decrypt(app_key.bytes, nonce, transport_pdu[1:])
        return AccessMessage(src, dst, ttl, decrypted_access)


class ControlMessage(Segment):
    def __init__(self, src, dst, ttl, opcode, payload):
        super().__init__(src, dst, ttl, True, payload)
        self.payload = payload
        self.opcode = opcode

    def get_opcode(self, application_key):
        return bitstring.pack("uint:7", self.opcode)

    def segments(self, application_key, seq, iv_index, szmic=False):
        yield from super()._segments(
            application_key, seq, payload=self.payload, szmic=False
        )

    @classmethod
    def decrypt(cls, ttl, src, dst, transport_pdu):
        seg, opcode = bitstring.BitString(transport_pdu).unpack("uint:1, uint:7")

        # works only for unsegmented messages!
        if seg:
            raise NotImplementedError
        return ControlMessage(src, dst, ttl, opcode, transport_pdu[1:])


class ProxyConfigMessage(Segment):
    def __init__(self, src, opcode, payload):
        super().__init__(src, 0x0000, 0x00, True, payload)
        self.payload = payload
        self.opcode = opcode

    def get_opcode(self, application_key):
        return bitstring.pack("uint:7", self.opcode)

    def segments(self, application_key, seq, iv_index, szmic=False):
        yield from super()._segments(
            application_key, seq, payload=self.payload, szmic=False
        )

    @classmethod
    def decrypt(cls, src, transport_pdu):
        opcode = transport_pdu[0]
        return ProxyConfigMessage(src, opcode, transport_pdu[1:])


class SolicitationMessage(Segment):
    def __init__(self, src, dst=0):
        super().__init__(src, dst, 0x00, True, bytes())
        self.payload = bytes()
        self.opcode = bytes()

    def get_opcode(self, application_key):
        return bitstring.BitString()

    def segments(self, application_key, seq, iv_index, szmic=False):
        yield bytes()

    @classmethod
    def decrypt(cls, src):
        return SolicitationMessage(src)


class SegmentAckMessage(ControlMessage):
    def __init__(self, src, dst, ttl, seq_zero, ack_segments, obo=False):
        self.obo = obo
        self.seq_zero = seq_zero
        self.block_ack = bitstring.BitArray(i in ack_segments for i in range(32))

        self.payload = bitstring.pack(
            "uint:1, uint:13, pad:2, bits:32",
            self.obo,
            self.seq_zero,
            reversed(self.block_ack),
        ).bytes

        super().__init__(src, dst, ttl, 0x00, self.payload)


class NetworkMessage:
    def __init__(self, message: Segment):
        self.message = message

    def pack(
        self,
        application_key,
        network_key,
        seq,
        iv_index,
        *,
        transport_seq=None,
        skip_segments=(),
    ):
        nid, encryption_key, privacy_key = network_key.encryption_keys

        # when retrying a segment, use the original sequence number during application
        # encryption, but a newer one on network layer
        if transport_seq is None:
            transport_seq = seq

        # remove segments that were ack-ed and encrypt with new network sequence
        segments = [
            segment
            for segment in self.message.segments(
                application_key, transport_seq, iv_index
            )
        ]
        for index in sorted(skip_segments, reverse=True):
            segments.pop(index)

        for seq, pdu in enumerate(segments, start=seq):
            if isinstance(self.message, ProxyConfigMessage):
                nonce = self.message.nonce.proxy(seq, iv_index)
            elif isinstance(self.message, SolicitationMessage):
                assert iv_index == 0x00000000
                nonce = self.message.nonce.solicitation(seq)
            else:
                nonce = self.message.nonce.network(seq, iv_index)
            network_pdu = aes_ccm_encrypt(
                encryption_key,
                nonce,
                bitstring.pack("uintbe:16, bytes", self.message.dst, pdu).bytes,
                b"",
                8 if self.message.ctl else 4,
            )

            network_header = bitstring.pack(
                "uint:1, uint:7, uintbe:24, uintbe:16",
                self.message.ctl,
                self.message.ttl,
                seq,
                self.message.src,
            ).bytes

            privacy_random = bitstring.pack(
                "pad:40, uintbe:32, bytes:7", iv_index, network_pdu[:7]
            ).bytes

            pecb = aes_ecb(privacy_key, privacy_random)[:6]

            obfuscated_header = bytes(map(operator.xor, network_header, pecb))

            yield seq, bitstring.pack(
                "uint:1, uint:7, bits, bytes",
                iv_index & 1,
                nid,
                obfuscated_header,
                network_pdu,
            ).bytes

    @classmethod
    def unpack(
        cls,
        app_key: ApplicationKey,
        net_key: NetworkKey,
        local_iv_index: int,
        network_pdu: bytes,
        proxy=False,
    ):
        # pylint: disable=R0914
        _nid, encryption_key, privacy_key = net_key.encryption_keys
        last_iv, nid, obfuscated_header, encoded_data_mic = bitstring.BitString(
            network_pdu
        ).unpack("uint:1, uint:7, bytes:6, bytes")
        if nid != _nid:
            raise KeyError
        iv_index = (
            local_iv_index if (local_iv_index & 0x01) == last_iv else local_iv_index - 1
        )
        privacy_random = bitstring.pack(
            "pad:40, uintbe:32, bytes:7", iv_index, encoded_data_mic[:7]
        ).bytes

        pecb = aes_ecb(privacy_key, privacy_random)[:6]
        deobfuscated = bytes(map(operator.xor, obfuscated_header, pecb))
        ctl, ttl, seq, src = bitstring.BitString(deobfuscated).unpack(
            "uint:1, uint:7, uintbe:24, uintbe:16"
        )
        net_mic_len = 8 if ctl else 4

        nonce = (
            Nonce(src, 0, ttl, ctl).proxy if proxy else Nonce(src, 0, ttl, ctl).network
        )(seq, iv_index)
        decrypted_net = aes_ccm_decrypt(
            encryption_key, nonce, encoded_data_mic, tag_length=net_mic_len
        )

        dst, transport_pdu = bitstring.BitString(decrypted_net).unpack(
            "uintbe:16, bytes"
        )

        if proxy:
            transport_msg = ProxyConfigMessage.decrypt(src, transport_pdu)
        elif ctl:
            transport_msg = ControlMessage.decrypt(ttl, src, dst, transport_pdu)
        else:
            transport_msg = AccessMessage.decrypt(
                app_key, iv_index, ctl, ttl, seq, src, dst, transport_pdu
            )
        net_message = NetworkMessage(transport_msg)
        return iv_index, seq, net_message


MESH_CRC = Configuration(
    width=8,
    polynomial=0x07,
    init_value=0xFF,
    final_xor_value=0xFF,
    reverse_input=True,
    reverse_output=True,
)

CRC_CALCULATOR = CrcCalculator(configuration=MESH_CRC)


class GenericProvisioningPDU:
    @staticmethod
    def pack(payload):
        if payload["type"] == ProvisioningPDUType.ACK:
            return [b"\x01"]

        if isinstance(payload["type"], BearerOpcode):
            return [
                ProvisioningBearerControl.build(
                    dict(opcode=payload["type"], parameters=payload["parameters"])
                )
            ]

        PDU = ProvisioningPDU.build(obj=payload)

        segments = [PDU[0:20]]
        if len(PDU) > 20:
            segments += [PDU[0 + i : 23 + i] for i in range(20, len(PDU), 23)]

        total_len = len(PDU)
        fcs = CRC_CALCULATOR.calculate_checksum(PDU)

        ret = list()
        ret.append(
            TransactionStartPDU.build(
                dict(
                    last_segment_number=len(segments) - 1,
                    total_length=total_len,
                    frame_check=fcs,
                    data=segments.pop(0),
                )
            )
        )

        for index, segment in enumerate(segments, start=1):
            ret.append(
                TransactionContinuationPDU.build(
                    dict(segment_index=index, data=segment)
                )
            )

        return ret

    @staticmethod
    def unpack(segments):
        if segments[0] == b"\x01":
            return dict(type=ProvisioningPDUType.ACK, parameters=dict())

        parsed = [TransactionPDUSegment.parse(segment) for segment in segments]
        parsed.sort(key=lambda segment: getattr(segment, "segment_index", 0))

        if parsed[0].get("opcode") is not None:
            return dict(type=parsed[0]["opcode"], parameters=parsed[0]["parameters"])

        return ProvisioningPDU.parse(data=b"".join(segment.data for segment in parsed))
